'''Tests for wrangler.py'''
import glob
import json
import os
from pathlib import Path
from typing import Dict

from numpy.random import RandomState
import pandas as pd
import pytest

from ..aggregators.impl.aggregator import AggregatorStub
from ..algorithms.impl.algorithm import Algorithm
from ..algorithms.impl.algorithm_auto import FakeAlgorithmCatalogAuto
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..algorithms.impl.fake_algorithm import FakeInstance
from ..executor.cucumber import Cucumber
from ..executor.executor import ExecutorStub
from ..executor.simple.simple_executor import SimpleExecutor
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (ExecutablePipeline,
                                                ExecutablePipelineStub,
                                                PipelineResult)
from ..problem_def.problem_def import ProblemDefinition
from ..templates.impl.pipeline_step import PipelineStep
from ..templates.impl.pipeline_template import PipelineTemplate
from ..templates.impl.template import TemplateCatalog

from .constants import Defaults
from .dataset import Dataset, RoleName
from .wrangler import Wrangler, WranglerFailure

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,duplicate-code,protected-access,too-many-ancestors
# pylint: disable=missing-class-docstring,redefined-outer-name


class FakeExecutor(ExecutorStub):

    def predict(self,
                dataset: Dataset,
                pipelines: Dict[Designator, ExecutablePipeline]
                ) -> Dict[Designator, PipelineResult]:
        '''Use a list of pipelines to predict from a dataset.'''
        predictions = dataset.output()
        # create a dataframe with the shape we expect predictions to be:
        # 1 column (the target) and a number of rows equal to the size of the input
        predictions.predictions = pd.DataFrame({
            dataset.metadata.roles[RoleName.TARGET][0].name:
                range(0, dataset.dataframe.shape[0])})
        return {d: PipelineResult(executable_pipeline=p, prediction=predictions)
                for d, p in pipelines.items()}


MEMORY_PROBLEM_DEF = '''
{
    "dataset": {
        "config": "memory",
        "column_roles": {
            "target": {
                "name": "c"
            }
        }
    },
    "problem_type": {
        "task": "regression"
    },
    "cross_validation": {
        "k": 10
    },
    "metrics" : {
        "root_mean_squared_error" : {}
    },
    "hyperparams": ["disable_grid_search"]
}
'''

AGGREGATION_PROBLEM_DEF = '''
{
    "dataset": {
        "config": "memory",
        "column_roles": {
            "target": {
                "name": "c"
            }
        }
    },
    "problem_type": {
        "task": "regression"
    },
    "cross_validation": {
        "k": 10
    },
    "metrics": {
        "root_mean_squared_error" : {},
        "mean_squared_error": {},
        "mean_absolute_error": {}
    },
    "aggregation": {
        "method": ["method1", "method2"]
    },
    "hyperparams": ["disable_grid_search"]
}
'''


def memory_problem_def(has_test_data: bool = False,
                       has_aggregation: bool = False) -> ProblemDefinition:
    data = pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(1, 21)
        }
    )
    test_data = None
    if has_test_data:
        test_data = data

    problem_def = MEMORY_PROBLEM_DEF
    if has_aggregation:
        problem_def = AGGREGATION_PROBLEM_DEF

    return ProblemDefinition(
        problem_def=problem_def,
        train_df=data,
        test_df=test_data)


def test_init_default() -> None:
    '''Test the components that wrangler defaults to.'''
    dut = Wrangler(problem_definition=memory_problem_def())
    assert dut._generator.__class__.__name__ == "GeneratorImpl"
    assert dut._template_catalog.__class__.__name__ == "TemplateCatalogAuto"


def test_init_custom() -> None:
    '''Test that overriding components when initializing the wrangler works.'''
    dut = Wrangler(
        problem_definition=memory_problem_def(),
        executor=FakeExecutor,
        template_catalog=FakeTemplateCatalog)
    assert dut._executor.__class__.__name__ == "FakeExecutor"
    assert dut._template_catalog.__class__.__name__ == "FakeTemplateCatalog"

    # Testing that non-set arguments are still default
    assert dut._generator.__class__.__name__ == "GeneratorImpl"


def test_wrangler_dataset() -> None:
    '''Test wrangler.dataset(), which takes a dataframe and returns a
    dataset containing the dataframe with metadata matching the problem definition.'''
    problem_def = memory_problem_def()
    metadata = problem_def.metadata
    data = {
        'a': range(1, 31),
        'b': range(1, 31),
        'c': range(1, 31),
    }

    dut = Wrangler(
        problem_definition=problem_def,
        executor=FakeExecutor)

    got = dut.dataset(data)

    # Confirm that the dataframe matches
    pd.testing.assert_frame_equal(
        got.dataframe,
        pd.DataFrame(data))

    # Confirm that at least part of the metadata matches.
    assert got.metadata.roles == metadata.roles


def test_wrangler_split() -> None:
    problem_def = memory_problem_def()
    dut = Wrangler(
        problem_definition=problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto,
        executor=FakeExecutor)

    train_result = dut.fit_predict_rank()

    # we expect 10 folds because k=10 was set in the problem definition
    assert len(train_result.split_dataset.folds) == 10
    # we expect each fold's validate set to have 2 rows,
    #    because there are 10 folds and the dataset is 20 rows long,
    #    and 2 columns, because it contains the attributes but not the target.
    assert train_result.split_dataset.folds[0].validate.dataframe.shape == (2, 2)
    # we expect each fold's train set to have 18 rows,
    #    because it is composed of 9 folds of 2 rows each,
    #    and 3 columns, becase it contains the attributes and the target.
    assert train_result.split_dataset.folds[0].train.dataframe.shape == (18, 3)
    # we expect the ground truth to have 20 rows,
    #    because it covers the whole dataset,
    #    and 1 col, because it only has the target.
    assert train_result.split_dataset.ground_truth is not None
    assert train_result.split_dataset.ground_truth.ground_truth.shape == (20, 1)


def test_fit_predict_rank() -> None:
    problem_def = memory_problem_def(has_test_data=True)

    dut = Wrangler(
        problem_definition=problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)

    got = dut.fit_predict_rank()

    # we expect pipelines to be named according to thier template (tabular regression)
    #    and the model that was chosen from thier query step.
    #    in the case of random forest regressor, wrangler will do grid search,
    #    and thus hyperparam values will also be in the designator.
    random_forest_des = Designator('tabular_regression@sklearn.ensemble.randomforestregressor:'
                                   'max_depth=None,min_samples_split=2')
    linear_regression_des = Designator('tabular_regression@sklearn.linear_model.linearregression')
    assert {random_forest_des, linear_regression_des} == set(got.executable_pipelines.keys())

    assert len(got.train_results) == 2  # 2 pipelines
    assert got.train_results.predictions[random_forest_des].predictions.shape == (20, 2)

    assert len(got.rankings) == 1  # 1 metric
    assert 'root_mean_squared_error' in got.rankings

    assert got.test_results is not None
    assert len(got.test_results) == 2  # 2 pipelines

    # executable pipelines associated with the test results should be identical
    assert set(got.executable_pipelines.keys()) == set(got.test_results.executable_pipelines.keys())

    assert got.test_results.predictions[random_forest_des].predictions.shape == (20, 1)


def test_fit_predict_rank_with_aggregation() -> None:
    problem_def = memory_problem_def(has_test_data=True, has_aggregation=True)

    dut = Wrangler(
        problem_definition=problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)

    dut.aggregator_catalog.register(AggregatorStub(name="method1"))
    dut.aggregator_catalog.register(AggregatorStub(name="method2"))

    got = dut.fit_predict_rank()

    assert len(got.rankings) == 5  # 3 metrics, and 2 aggregate metrics
    want = {
        'root_mean_squared_error', 'mean_squared_error', 'mean_absolute_error',
        'method1', 'method2'
    }
    assert want == set(got.rankings.keys())


class FakeTemplateCatalog(TemplateCatalog):
    '''fake, no templates'''


def memory_test_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            'a': range(1, 11),
            'b': range(1, 11)
        }
    )


def memory_train_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(1, 21)
        }
    )


def test_predict() -> None:
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    alg_catalog = FakeAlgorithmCatalogAuto
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor,
        algorithm_catalog=alg_catalog)
    got = dut.fit_predict_rank()

    fake_des = Designator('fake_des')

    testset = dut.dataset(memory_test_df())
    result = list(got.train_results.values())[0]
    pipeline = result.executable_pipeline
    assert pipeline is not None
    test_got = dut.predict(
        new_data=testset,
        trained_pipelines={fake_des: pipeline})
    prediction_df = test_got[fake_des].prediction.predictions

    # we expect 10 rows (length of test dataframe) and 1 col (target)
    assert prediction_df.shape == (10, 1)
    # we expect to be able to predict perfectly since test data is identical to train data.
    assert prediction_df['c'][9] == 10.0


def test_predict_defaulted() -> None:
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    alg_catalog = FakeAlgorithmCatalogAuto
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor,
        algorithm_catalog=alg_catalog)
    _ = dut.fit_predict_rank()

    testset = dut.dataset(memory_test_df())
    test_got = dut.predict(
        new_data=testset)
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=None,min_samples_split=2',
    } == set(test_got.keys())


def test_set_current_predict() -> None:
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    alg_catalog = FakeAlgorithmCatalogAuto
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor,
        algorithm_catalog=alg_catalog)
    _ = dut.fit_predict_rank()

    testset = dut.dataset(memory_test_df())

    # Pick one pipeline.
    dut.set_current('tabular_regression@sklearn.ensemble.randomforestregressor:'
                    'max_depth=None,min_samples_split=2')
    test_got = dut.predict(new_data=testset)
    assert set(test_got.keys()) == {
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=None,min_samples_split=2',
    }

    # Reset to all known pipelines.
    dut.set_current()

    test_got = dut.predict(new_data=testset)
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=None,min_samples_split=2',
    } == set(test_got.keys())


def test_set_current_fit() -> None:
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    alg_catalog = FakeAlgorithmCatalogAuto
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor,
        algorithm_catalog=alg_catalog)
    _ = dut.fit_predict_rank()

    testset = dut.dataset(memory_train_df())

    # Pick one pipeline.
    dut.set_current('tabular_regression@sklearn.ensemble.randomforestregressor:'
                    'max_depth=None,min_samples_split=2')

    test_got = dut.fit(dataset=testset)
    assert set(test_got.keys()) == {
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=None,min_samples_split=2',
    }

    # Reset to all known pipelines.
    dut.set_current()

    test_got = dut.fit(dataset=testset)
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=None,min_samples_split=2',
    } == set(test_got.keys())


def test_predict_bad_executor_kind() -> None:
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor)

    fake_des = Designator('fake_des')
    stub_pipeline = ExecutablePipelineStub(trained=True)

    with pytest.raises(WranglerFailure, match=r'(fake_des.*simple)|(simple.*fake_des)/i'):
        dut.predict(
            new_data=Dataset(),
            trained_pipelines={fake_des: stub_pipeline})


def test_predict_untrained_pipeline() -> None:
    problem_def = memory_problem_def()
    executor = FakeExecutor  # executor kind is 'stub_executor_kind', should match stub pipeline.
    dut = Wrangler(
        problem_definition=problem_def,
        executor=executor)

    fake_des = Designator('fake_des')
    stub_pipeline = ExecutablePipelineStub(trained=False)

    with pytest.raises(WranglerFailure, match=r'(fake_des.*trained)|(trained.*fake_des)/i'):
        dut.predict(
            new_data=Dataset(),
            trained_pipelines={fake_des: stub_pipeline})


@pytest.fixture(scope='session')
def saver_problem_def(tmp_path_factory: pytest.TempPathFactory) -> ProblemDefinition:
    tmp_path = tmp_path_factory.mktemp('data')
    tmp_dir = tmp_path / 'sub'
    tmp_dir.mkdir()
    config = {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            }
        },
        "problem_type": {
            "task": "regression"
        },
        "cross_validation": {
            "k": 10
        },
        "metrics": {
            "root_mean_squared_error": {}
        },
        "output": {
            "path": tmp_dir,
            "instantiations": [
                "json"
            ],
            "file_type": "csv"
        },
        "hyperparams": ["disable_grid_search"]
    }
    data = pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(1, 21)
        }
    )

    return ProblemDefinition(
        problem_def=config,
        train_df=data)


def test_save_filesystem(saver_problem_def: ProblemDefinition) -> None:
    PipelineStep.reset_serial_number()
    dut = Wrangler(
        problem_definition=saver_problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)
    _ = dut.fit_predict_rank()

    path = saver_problem_def.output.path
    assert path is not None
    filenames = glob.glob(str(path / 'models' / '*.linearregression' / '@connect_*@.pkl'))
    assert len(filenames) == 1
    with Path(filenames[0]).open(mode='rb') as filepointer:
        cucumber = Cucumber.deserialize(filepointer.read())
    assert cucumber.hyperparams == {
        'target': ['target', 'dataframe'],
        'covariates': ['attributes', 'dataframe']
    }

    pipelines = os.listdir(str(path / 'pipelines'))
    assert {'tabular_regression@sklearn.ensemble.randomforestregressor:'
            'max_depth=none,min_samples_split=2.json',
            'tabular_regression@sklearn.linear_model.linearregression.json',
            } == set(pipelines)

    with (path / 'pipelines' / 'tabular_regression@sklearn.linear_model.linearregression.json'
          ).open() as file_pointer:
        got = json.load(file_pointer)

    assert got['pipeline_designator'] == 'tabular_regression@sklearn.linear_model.linearregression'


@pytest.fixture(scope='session')
def search_problem_def() -> ProblemDefinition:
    config = '''
    {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            }
        },
        "problem_type": {
            "task": "regression"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics" : {
            "root_mean_squared_error" : {}
        },
        "hyperparams": [
            {
                "select": {
                    "algorithm": "sklearn.ensemble.RandomForestRegressor"
                },
                "params": {
                    "n_estimators": {
                        "list": [100, 200]
                    }
                }
            }
        ]
    }
    '''
    data = pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(1, 21)
        }
    )

    return ProblemDefinition(
        problem_def=config,
        train_df=data)


def test_hyperparam_search(search_problem_def: ProblemDefinition) -> None:
    dut = Wrangler(
        problem_definition=search_problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)
    got = dut.fit_predict_rank()
    assert len(got.executable_pipelines) == 25
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=8,min_samples_split=10,n_estimators=100',
        'tabular_regression@sklearn.ensemble.randomforestregressor:'
        'max_depth=15,min_samples_split=10,n_estimators=200',
    }.issubset(got.executable_pipelines.keys())


@pytest.fixture(scope='session')
def order_problem_def() -> ProblemDefinition:
    config = '''
    {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            }
        },
        "problem_type": {
            "task": "test_task"
        },
        "cross_validation": {
            "k": 3
        },
        "metrics" : {
            "root_mean_squared_error" : {}
        },
        "hyperparams": ["disable_grid_search"]
    }
    '''
    data = pd.DataFrame(
        {
            'a': range(1, 22),
            'b': range(1, 22),
            'c': range(2, 23)
        }
    )

    return ProblemDefinition(
        problem_def=config,
        train_df=data)


class FakeInst(FakeInstance):
    '''The instance of a fake algorithm for order tests.'''
    def predict(self, dataset: Dataset) -> Dataset:
        retval = dataset.output()
        df = pd.DataFrame({'c': dataset.dataframe['a'] + 1})
        retval.predictions = df
        return retval


class FakeAlg(Algorithm):
    '''A fake algorithm used for order tests.'''
    _name = 'fake_algorithm'

    def instantiate(self, **hyperparams) -> 'FakeInstance':
        return FakeInst(parent=self, **hyperparams)


def test_order(order_problem_def) -> None:
    dut = Wrangler(
        problem_definition=order_problem_def)

    fake_template = PipelineTemplate(
        name='test_template',
        tags={
            'task': ['test_task'],
            'data_type': ['tabular']},
        generator=dut.generator
    )
    fake_template.step(model=FakeAlg())
    dut.template_catalog.register(fake_template)

    got = dut.fit_predict_rank()
    predict_df = got.train_results.predictions[Designator('test_template')].predictions
    assert got.split_dataset.ground_truth is not None
    ground_truth_df = got.split_dataset.ground_truth.ground_truth
    pd.testing.assert_frame_equal(predict_df[['c']], ground_truth_df)
    assert all(predict_df['c'] == predict_df['ground truth'])


def non_seeded_problem_def() -> ProblemDefinition:
    config = '''
    {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            }
        },
        "problem_type": {
            "task": "test_task"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics" : {
            "root_mean_squared_error" : {}
        },
        "hyperparams": ["disable_grid_search"]
    }
    '''
    data = pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(2, 22)
        }
    )

    return ProblemDefinition(
        problem_def=config,
        train_df=data)


def seeded_problem_def(seed: int) -> ProblemDefinition:
    config = {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            }
        },
        "problem_type": {
            "task": "test_task"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics": {
            "root_mean_squared_error": {}
        },
        "hyperparams": [
            "disable_grid_search",
            {
                "select": {
                    "tags": {
                        "supports_random_seed": "true"
                    }
                },
                "params": {
                    "random_seed": {
                        "fixed": seed
                    }
                }
            }
        ]
    }
    data = pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(2, 22)
        }
    )

    return ProblemDefinition(
        problem_def=config,
        train_df=data)


class RandomAlgorithmInstance(AlgorithmInstance):
    _rng: RandomState

    def __init__(self, parent: Algorithm, **hyperparams):
        super().__init__(parent)
        if 'random_seed' in hyperparams:
            self._rng = RandomState(hyperparams['random_seed'])

    def predict(self, dataset: Dataset) -> Dataset:
        retval = dataset.output()
        retval['predictions'] = pd.DataFrame({
            'c': list(self._rng.randint(0, 10, len(dataset.dataframe)))
        }).reset_index(drop=True)
        return retval


class RandomAlgorithm(Algorithm):
    _name = 'random_algorithm'
    _tags = {
        'task': ['test_task'],
        'data_type': ['tabular'],
        'supports_random_seed': ['true']
    }
    _default_hyperparams = {
        'random_seed': Defaults.SEED,
    }

    def instantiate(self, **hyperparams) -> RandomAlgorithmInstance:
        return RandomAlgorithmInstance(parent=self, **self.hyperparams(**hyperparams))


def test_set_seed() -> None:
    dut_def = Wrangler(problem_definition=non_seeded_problem_def())
    dut5678 = Wrangler(problem_definition=seeded_problem_def(5678))

    pipe_def = PipelineTemplate(
        name='test_template',
        tags={
            'task': ['test_task'],
            'data_type': ['tabular']},
        generator=dut_def.generator
    )
    pipe_def.step(model=RandomAlgorithm())
    dut_def.template_catalog.register(pipe_def)

    pipe5678 = PipelineTemplate(
        name='test_template',
        tags={
            'task': ['test_task'],
            'data_type': ['tabular']},
        generator=dut5678.generator
    )
    pipe5678.step(model=RandomAlgorithm())
    dut5678.template_catalog.register(pipe5678)

    got_def = dut_def.fit_predict_rank()
    got5678 = dut5678.fit_predict_rank()

    des_def = Designator('test_template')
    des5678 = Designator('test_template@random_algorithm:random_seed=5678')

    predictions_def = got_def.train_results[des_def].prediction.predictions[['c']]
    predictions5678 = got5678.train_results[des5678].prediction.predictions[['c']]

    # The k-fold-cross-validator instantiates the algorithm separately
    # for each fold, so in this special case, we expect to see the same
    # sequence twice.

    want_def = pd.DataFrame({'c': [3, 3, 6, 2, 1, 8, 5, 1, 7, 3, 3, 3, 6, 2, 1, 8, 5, 1, 7, 3]})
    want5678 = pd.DataFrame({'c': [5, 3, 6, 9, 3, 3, 9, 8, 3, 3, 5, 3, 6, 9, 3, 3, 9, 8, 3, 3]})
    pd.testing.assert_frame_equal(predictions_def, want_def)
    pd.testing.assert_frame_equal(predictions5678, want5678)


@pytest.fixture(scope='session')
def disable_grid_search_problem_def() -> ProblemDefinition:
    config = '''
    {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            }
        },
        "problem_type": {
            "task": "regression"
        },
        "cross_validation": {
            "k": 2
        },
        "metrics" : {
            "root_mean_squared_error" : {}
        },
        "hyperparams": [
            "disable_grid_search",
            {
                "select": {
                    "algorithm": "sklearn.ensemble.RandomForestRegressor"
                },
                "params": {
                    "n_estimators": {
                        "list": [100, 200],
                        "default": 100
                    }
                }
            }
        ],
        "output": {}
    }
    '''
    data = pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(1, 21)
        }
    )

    return ProblemDefinition(
        problem_def=config,
        train_df=data)


def test_disable_grid_search(disable_grid_search_problem_def: ProblemDefinition) -> None:
    dut = Wrangler(
        problem_definition=disable_grid_search_problem_def,
        algorithm_catalog=FakeAlgorithmCatalogAuto)
    got = dut.fit_predict_rank()
    assert {
        'tabular_regression@sklearn.linear_model.linearregression',
        'tabular_regression@sklearn.ensemble.randomforestregressor'
        ':max_depth=None,min_samples_split=2,n_estimators=100'
    } == set(got.executable_pipelines.keys())



#!/bin/python
"""Main for AutonML."""

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import argparse
import logging
import sys
import warnings
from typing import Callable

from ngautonml.algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ngautonml.executor.simple.simple_executor import SimpleExecutor
from ngautonml.generator.generator import GeneratorImpl
from ngautonml.instantiator.instantiator_factory import InstantiatorFactory
from ngautonml.metrics.impl.metric_auto import MetricCatalogAuto
from ngautonml.problem_def.problem_def import ProblemDefinition
from ngautonml.ranker.ranker_impl import RankerImpl
from ngautonml.templates.impl.template_auto import TemplateCatalogAuto
from ngautonml.wrangler.wrangler import Wrangler

logging.basicConfig(level=logging.ERROR)
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)


def wrangle(args: argparse.Namespace):
    """Do full autoML."""
    # TODO(piggy): Set up non-trivial architectural componenents
    # here based on command-line arguments.
    with open(args.problem_definition, 'r', encoding='utf8') as _f:
        problem_definition = ProblemDefinition(_f.read())
    logging.info('%s', problem_definition)
    wrangler = Wrangler(
        problem_definition=problem_definition,
        metric_catalog=MetricCatalogAuto,
        algorithm_catalog=AlgorithmCatalogAuto,
        ranker=RankerImpl,
        template_catalog=TemplateCatalogAuto,
        generator=GeneratorImpl,
        executor=SimpleExecutor,
        instantiator_factory=InstantiatorFactory,
    )
    got = wrangler.fit_predict_rank()
    print(got.rankings)


def instantiate(args: argparse.Namespace):
    """Instantiate one or more pipelines.

    Optionally includes a trained model in the instantiated pipeline.
    """
    print(f'Instantiate: {args!r}')


def usage(parser: argparse.ArgumentParser) -> Callable[
        [argparse.Namespace], None]:
    """Returns a function that prints the usage message."""
    def _f(_args: argparse.Namespace):
        parser.print_usage(sys.stderr)
    return _f


def main():
    """
    Just do something
    """
    # Create the top-level parser.
    parser_global = argparse.ArgumentParser(prog='picard')
    parser_global.set_defaults(func=usage(parser_global))

    # Create sub-parser.
    sub_parsers = parser_global.add_subparsers(help='sub-command help')

    # Create the parser for the "wrangle" sub-command.
    parser_wrangle = sub_parsers.add_parser('wrangle',
                                            help='Run full autoML on a '
                                            'dataset.')
    parser_wrangle.set_defaults(func=wrangle)
    parser_wrangle.add_argument(
        '-d', '--problem_definition',
        type=str,
        help='path to problem definition JSON file',
        required=True)

    # Create the parser for the "instantiate" sub-command.
    parser_instantiate = sub_parsers.add_parser('instantiate',
                                                help='Generate an '
                                                'executable algorithm.')
    parser_instantiate.set_defaults(func=instantiate)
    parser_instantiate.add_argument(
        '-k', '--kind',
        action='append', choices=['d3m_json', 'python_script',
                                  'jupyter_notebook', 'airflow_class'],
        help='Indicate the kind of executable algorithm to generate. '
        'May be repeated. Defaults to all choices.')
    parser_instantiate.add_argument(
        '-p', '--pipeline',
        type=str,
        help='name or path to a pipeline to instantiate',
        required=True
    )
    parser_instantiate.add_argument(
        '-m', '--trained_model',
        action='append',
        help='optional path to the trained model to instantiate. '
        'May be repeated.'
    )

    a = parser_global.parse_args()
    a.func(a)


if __name__ == "__main__":
    main()



'''This is the executable pipeline for the "simple" executor.'''
from typing import List, Optional

from ...algorithms.impl.fittable_algorithm_instance import UntrainedError
from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import Designator
from ...instantiator.executable_pipeline import (ExecutablePipeline,
                                                 PipelineResult,
                                                 FitError)
from ...templates.impl.parallel_step import ParallelStep
from ...wrangler.dataset import Dataset, DatasetKeys
from ..cucumber import JarOfCucumbers
from ..executor_kind import ExecutorKind
from .simple_executable_step import SimpleExecutableStep
from .simple_parallel_executable_step import SimpleParallelExecutableStep

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class SimpleExecutablePipeline(ExecutablePipeline):
    '''A pipeline for the 'simple' executor.'''
    _kind = ExecutorKind('simple')
    _steps: List[SimpleExecutableStep]

    def __init__(self,
                 pipeline: BoundPipeline):
        trained = False
        self._pipeline = pipeline
        self._steps = []
        for step in pipeline.steps:
            if isinstance(step, ParallelStep):
                executable_step: SimpleExecutableStep = (
                    SimpleParallelExecutableStep(step, SimpleExecutablePipeline))
            else:
                executable_step = SimpleExecutableStep(step)
                if executable_step.trained:
                    trained = True
            self._steps.append(executable_step)
        if trained:
            # Mainly, this catches the subpipelines of parallel steps
            # which are not otherwise marked as trained.
            self.set_trained()

    def set_trained(self):
        super().set_trained()
        for step in self._steps:
            step.set_trained()

    def set_fit_error(self, err: FitError) -> None:
        '''Set this pipeline as throwing an error at fit time'''
        self._fit_error = err

    def fit(self, dataset: Dataset) -> JarOfCucumbers:
        '''Fit models to dataset.

        Args:
            dataset: The training dataset.

        Returns:
            A JarOfCucumbers
            (mapping from step designators to cucumbers for each step).
        '''
        # Pipelines set themselves as trained first, so that they are
        #   marked trained even if they throw an exception along the way
        self._trained = True

        inp = dataset
        for step in self._steps:
            step.fit(dataset=inp)
            result = step.predict(dataset=inp)
            inp = result
        return self.cucumberize_all()

    def predict(self, dataset: Dataset) -> PipelineResult:
        '''Run prediction for a complete pipeline.'''
        if not self._trained:
            raise UntrainedError(f'pipeline "{self.name}" needs to be fit before it can predict')
        inp = dataset.output()
        if self._fit_error is not None:
            inp[DatasetKeys.ERROR.value] = self._fit_error
        else:
            inp.update(dataset)
            for step in self._steps:
                result = step.predict(dataset=inp)
                inp = result
        return PipelineResult(executable_pipeline=self, prediction=inp)

    def cucumberize_all(self, parent_designator: Optional[Designator] = None) -> JarOfCucumbers:
        '''Cucumberize all steps that hold algorithms, and return the result.

        This has the same output as fit() but requires the pipeline to be fit already.

        The parent_designator argument is not None iff this pipeline is a subpipeline inside
            a parallel step.  It is there so that steps in subpipelines can be saved with
            their parent's designator.
        '''
        if not self._trained:
            raise UntrainedError(
                f'pipeline "{self.name}" needs to be fit before it can cucumberize_all')

        models = JarOfCucumbers()
        for step in self._steps:
            models.update(
                step.cucumberize_all(
                    parent_designator or self.designator))

        return models



'''Searcher object.

This object sits between the Generator and the Instantiator.
It implements static searching algoritms (like grid search)
and does hyperparameter overriding (a trivial special case
of search).
'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
import logging
from typing import Any, Dict, List, Optional

from frozendict import frozendict

from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..problem_def.hyperparam_config import HyperparamConfig
from ..templates.impl.pipeline_step import PipelineStep
from ..wrangler.constants import RangeMethod

from .frozen_overrides import FrozenOverrides
from .params import Overrides, ParamRanges
from .param_range import ParamRange


class Searcher(metaclass=abc.ABCMeta):
    '''Generate new BoundPipelines based on hyperparam search spaces.'''
    _overrides: Overrides
    _disable_grid_search: bool

    def __init__(
        self,
        hyperparams: HyperparamConfig
    ):
        self._overrides = hyperparams.overrides
        self._disable_grid_search = hyperparams.disable_grid_search

    @abc.abstractmethod
    def bind_hyperparams(self, pipeline: BoundPipeline) -> Dict[Designator, BoundPipeline]:
        '''Bind new BoundPipelines with revised hyperparameters.'''

    def bind_all(self,
                 pipelines: Dict[Designator, BoundPipeline]) -> Dict[Designator, BoundPipeline]:
        '''Bind all the hyperparams for a set of BoundPipelines.'''
        retval: Dict[Designator, BoundPipeline] = {}
        for pipe in pipelines.values():
            retval.update(self.bind_hyperparams(pipe))
        return retval


class TrivialSearcher(Searcher):
    '''Just return the pipeline passed to us. Not really a searcher.'''
    def bind_hyperparams(self, pipeline: BoundPipeline) -> Dict[Designator, BoundPipeline]:
        return {
            pipeline.designator: pipeline
        }


class StepSet():
    '''Turn a step into a set of steps for all hyperparam bindings.'''

    # The key of _steps is optional because at initialization we
    # need to use None as a key.
    _steps: Dict[Optional[frozendict[str, str]], PipelineStep]
    _algorithm: Any  # Breaking circularity; really an Algorithm
    _pipeline_designator_component: str

    def __init__(self, step: PipelineStep):
        self._steps = {None: step}
        assert step.has_model(), (
            'BUG: attempt to bind hyperparams for step '
            f'({step.pipeline_designator_component}) with no algorithm')
        self._algorithm = step.model
        self._pipeline_designator_component = step.pipeline_designator_component

    def multiply(self, hyperparam_name: str, param_range: ParamRange) -> 'StepSet':
        '''Expand the set of steps by the number of values in param.'''
        new_steps: Dict[Optional[frozendict[str, str]], PipelineStep] = {}
        for old_bindings, old_step in self._steps.items():
            bindings: Dict[str, Any] = self._algorithm.bind(
                hyperparam_name=hyperparam_name, param_range=param_range)
            for new_value_str, new_value in bindings.items():

                new_step = old_step.clone(**{hyperparam_name: new_value})
                assert new_step.filename is not None

                if old_bindings is None:
                    # step does not have any pre-existing overrides
                    new_bindings = frozendict({hyperparam_name: new_value_str})
                else:
                    # step has preexisting overrides
                    new_bindings = old_bindings.set(hyperparam_name, new_value_str)

                new_steps[new_bindings] = new_step
        self._steps = new_steps
        return self

    @property
    def steps(self) -> Dict[Optional[FrozenOverrides], PipelineStep]:
        '''Render self as a dict of PipelineSteps.'''
        retval: Dict[Optional[FrozenOverrides], PipelineStep] = {}
        for bind, step in self._steps.items():
            if bind is None:
                retval[None] = step
                continue
            retval[FrozenOverrides({
                self._pipeline_designator_component: bind
            })] = step
        return retval


class PipelineBuilder():
    '''Build a set of pipelines by repeatedly adding a single Step or a StepSet.'''
    _pipename: str
    _steps: Optional[Dict[Optional[FrozenOverrides], List[PipelineStep]]]

    def __init__(self, pipename: str):
        self._pipename = pipename
        self._steps = None

    def append(self, step_set: StepSet) -> 'PipelineBuilder':
        '''Expand self._steps, adding each step from step_set to the end of each step list.

        If len(self._steps) is n, and len(step_set) is m, len(self._steps) will become m*n.
        '''
        if self._steps is None:
            self._steps = {key: [step] for key, step in step_set.steps.items()}
            return self
        new_steps: Dict[Optional[FrozenOverrides], List[PipelineStep]] = {}
        for old_frozen_overrides, old_steps in self._steps.items():
            for new_frozen_overrides, new_step in step_set.steps.items():

                if new_frozen_overrides is not None and old_frozen_overrides is not None:
                    updated_frozen_overrides = old_frozen_overrides.update(
                        new_frozen_overrides.thaw())
                elif new_frozen_overrides is not None:
                    updated_frozen_overrides = new_frozen_overrides
                elif old_frozen_overrides is not None:
                    updated_frozen_overrides = old_frozen_overrides
                else:
                    # Case: both None
                    updated_frozen_overrides = FrozenOverrides()

                new_steps[updated_frozen_overrides] = old_steps + [new_step]

        self._steps = new_steps
        return self

    def append_step(self, step: PipelineStep) -> 'PipelineBuilder':
        '''Add a single step to every step list in self._steps.

        (used for steps that don't match any selectors because they have no hyperparam overrides)
        '''
        if self._steps is None:
            self._steps = {None: [step]}
            return self
        for key in self._steps:
            self._steps[key].append(step)
        return self

    def build(self) -> Dict[Designator, BoundPipeline]:
        '''Render the final BoundPipelines from lists of steps.'''
        retval: Dict[Designator, BoundPipeline] = {}
        if self._steps is None:
            return retval
        for frozen_overrides, rev_steps in self._steps.items():
            new_pipeline = BoundPipeline.build(
                steps=list(reversed(rev_steps)),
                template_name=self._pipename,
                frozen_overrides=frozen_overrides)
            retval[new_pipeline.designator] = new_pipeline
        return retval


class SearcherImpl(Searcher):
    '''This is the default searcher.'''

    def bind_hyperparams(self, pipeline: BoundPipeline) -> Dict[Designator, BoundPipeline]:
        '''Bind this Searcher's hyperparam range(s) to the pipeline'''
        builder = PipelineBuilder(pipeline.name)
        for step in reversed(pipeline.steps):
            default_alg_ranges = ParamRanges(
                {k: p for k, p in step.hyperparams().items() if isinstance(p, ParamRange)}
            )
            overrides = Overrides(self._overrides)

            params = default_alg_ranges.copy()
            for override in overrides:
                if override.selector.matches(step, pipeline):
                    logging.info('Matched: %s\n\tto step: %s\n\tin pipeline: %s\n',
                                 override,
                                 step.pipeline_designator_component,
                                 pipeline.designator)
                    params.update(override.params)

            if len(params) > 0:
                step_set = StepSet(step)
                for hyperparam_name, param_range in params.items():
                    range_to_multiply = param_range
                    if self._disable_grid_search:
                        # Replace existing range with fixed range at default
                        range_to_multiply = ParamRange(
                            method=RangeMethod.FIXED,
                            prange=param_range.default)
                    step_set.multiply(hyperparam_name, range_to_multiply)

                builder.append(step_set)
            else:
                builder.append_step(step)

        return builder.build()



'''Base class for elements of the Problem Definition file'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
import copy
from typing import Any, Dict, List, Optional, Union
from ..wrangler.constants import ProblemDefKeys


class ConfigError(BaseException):
    '''All errors thrown by problem definition modules'''


class ConfigDirectoryError(ConfigError, NotADirectoryError):
    '''A path given in the problem doc was not a directory, but should be.'''


class MissingKeyError(ConfigError, KeyError):
    '''We failed to find a given key in the problem doc.'''


class InvalidKeyError(ConfigError, KeyError):
    '''We found an invalid key in the problem doc.'''


class ProblemDefTypeError(ConfigError, TypeError):
    '''A value in the problem definition was the wrong type.'''


class InvalidValueError(ConfigError, ValueError):
    '''A value in the problem definition has the right type but is invalid.'''


class ParsingErrors(ConfigError):
    '''Error(s) thrown when problem definition JSON cannot be parsed'''

    def __init__(self, errors: List[ConfigError]):
        super().__init__(f'At least one parsing error occured: {errors!r}')
        self.errors = errors


class ValidationErrors(ConfigError):
    '''Problem definition JSON can be parsed, but is invalid in some other way.'''

    def __init__(self, errors: List[ConfigError]):
        super().__init__('At least one validation error occured:')
        self.errors = errors

    def __str__(self) -> str:
        retval_list1 = [str(arg) for arg in self.args]
        retval_list2 = [f'\t{repr(error)}' for error in self.errors]
        return '\n'.join(retval_list1 + retval_list2)


class ConfigComponent(metaclass=abc.ABCMeta):
    '''Base class for elements of the Problem Definition file

    This provides _get, _get_with_default, and _exists for
    use by config components.
    '''
    # pylint: disable=too-few-public-methods
    _parents: List[str]
    _clause: Dict[str, Any]

    def __init__(self, clause: Dict[str, Any], parents: Optional[List[str]] = None):
        if parents is None:
            parents = []
        self._parents = parents
        self._clause = clause

    def _add_parent(self,
                    parents: Optional[List[str]],
                    new_parent: str) -> List[str]:
        if parents is None:
            retval = []
        else:
            retval = parents.copy()
        retval.append(new_parent)
        return retval

    def _get(self, *args: Union[int, str, ProblemDefKeys]):
        '''for ex: call _get('metric', 'name') to get metric name'''
        here = self._clause
        so_far = []
        for k in args:
            if isinstance(k, ProblemDefKeys):
                k = k.value
            so_far.append(str(k))
            if isinstance(here, list):
                if not isinstance(k, int):
                    raise ProblemDefTypeError(
                        f'found a list at {".".join(self._parents + so_far[:-1])}, '
                        f'expected an integer at {".".join(self._parents + so_far)}')
                if not -len(here) < k < len(here):
                    raise MissingKeyError(f'Could not find the {k}th element of {here}')
            elif isinstance(here, dict):
                if k not in here:
                    raise MissingKeyError('Could not find '
                                          f'{".".join(so_far)}')
                assert isinstance(k, str), (
                    f'BUG: at _get {".".join(so_far)} we have a non-string key, {k!r} for {here!r}')
            here = here[k]

        return copy.deepcopy(here)

    def _get_with_default(self, *args: Union[int, str, ProblemDefKeys], dflt: Any) -> Any:
        if self._exists(*args):
            return self._get(*args)
        return dflt

    def _exists(self, *args: Union[int, str, ProblemDefKeys]) -> bool:
        '''for ex: call _exists('metric', 'name') to see if metric.name exists'''
        here = self._clause
        so_far = []
        for k in args:
            if isinstance(k, ProblemDefKeys):
                k = k.value
            so_far.append(str(k))
            if isinstance(here, list):
                if not isinstance(k, int):
                    raise ProblemDefTypeError(
                        f'found a list at {".".join(self._parents + so_far[:-1])}, '
                        f'expected an integer at {".".join(self._parents + so_far)}')
                if not -len(here) < k < len(here):
                    raise MissingKeyError(f'Could not find the {k}th element of {here}')
            elif isinstance(here, dict):
                if k not in here:
                    return False
                assert isinstance(k, str), (
                    f'BUG: at _exists {".".join(so_far)} we have a '
                    f'non-string key, {k!r} for {here!r}')
            here = here[k]
        return True

    @abc.abstractmethod
    def validate(self, **kwargs) -> None:
        '''Raise a ConfigError if something about the clause is invalid.'''



'''Tests for problem_def.py'''
from pathlib import Path

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd
import pytest

from ..dataset_formats.local import LocalDatasetConfig
from ..dataset_formats.memory import MemoryDatasetConfig
from ..wrangler.dataset import RoleName, Column
from .config_component import ParsingErrors
from .output_config import OutputConfigError
from .problem_def import ProblemDefinition, Task, ValidationErrors


# pylint: disable=missing-function-docstring,protected-access
# pylint: disable=redefined-outer-name


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    return tmp_path_factory.mktemp("data")


def module_path() -> Path:
    pathobj = Path(__file__).resolve()
    module_parent = pathobj.parents[2]
    return module_parent


TEST_PROBLEM_DEF = '''{
    "_comments" : [
        "A json file fully encapsulating the problem definition",
        "Specifically, an exmaple time series forecasting problem."
    ],
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "output": {}
}
'''


def test_get():
    dut = ProblemDefinition(TEST_PROBLEM_DEF)
    assert dut._get('problem_type', 'task') == 'regression'
    assert set(dut._get().keys()) == set(
        ['output', 'dataset', 'problem_type'])


def test_read_from_file(tmp_path: Path) -> None:
    tmp_file = tmp_path / 'pdef.json'
    with open(str(tmp_file), 'w', encoding='utf8') as f:
        f.write(TEST_PROBLEM_DEF)
    dut = ProblemDefinition(str(tmp_file))
    assert dut._get('problem_type', 'task') == 'regression'
    assert set(dut._get().keys()) == set(
        ['output', 'dataset', 'problem_type'])


def test_read_from_file_fail() -> None:
    '''We want to throw multiple errors for an invalid string.

    If a string is given that cannot be parsed as JSON or a filepath,
    we want to show the errors from both attempts to parse it.
    '''
    match_str = r'(JSONDecode.*FileNotFound)|(FileNotFound.*JSONDecode)'
    with pytest.raises(ParsingErrors,
                       match=match_str):
        ProblemDefinition('foo.foo')


def test_dataset_config() -> None:
    path = module_path() / 'examples' / 'classification' / 'credit.csv'
    pdef = {
        "dataset": {
            "config": "local",
            "train_path": path,
            "column_roles": {}
        },
        "problem_type": {
            "task": "regression"
        },
        "output": {}
    }
    dut = ProblemDefinition(pdef)
    assert isinstance(dut.dataset_config, LocalDatasetConfig)


MISSING_KEY = '''{
    "metrics" : {}
}
'''

# Missing "output" key


def test_missing_keys() -> None:
    with pytest.raises(ValidationErrors, match=r"[Mm]issing"):
        ProblemDefinition(MISSING_KEY)


INVALID_KEY = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "hamster": {}
}
'''


def test_invalid_keys() -> None:
    with pytest.raises(ValidationErrors, match=r"hamster"):
        ProblemDefinition(INVALID_KEY)


METRIC_MULTI = '''{
    "metrics" : {
        "metric1" : {
            "catalog_name": "real_metric_name",
            "k1" : "v1"
        },
        "metric2" : {
            "catalog_name": "real_metric_name",
            "k1" : "v2"
        },
        "another_metric": {}
    },
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    }
}
'''


def test_metric_sunny_day() -> None:
    dut = ProblemDefinition(METRIC_MULTI).metric_configs
    assert len(dut) == 3
    dutm1 = dut['metric1']
    assert dutm1.name == 'metric1'
    assert dutm1.hyperparams == {'k1': 'v1'}
    assert dutm1.catalog_name == 'real_metric_name'
    dutm2 = dut['metric2']
    assert dutm2.name == 'metric2'
    assert dutm2.hyperparams == {'k1': 'v2'}
    assert dutm2.catalog_name == 'real_metric_name'
    dutanother = dut['another_metric']
    assert dutanother.name == 'another_metric'
    assert dutanother.hyperparams == {}
    assert dutanother.catalog_name == 'another_metric'


def test_no_metrics() -> None:
    dut = ProblemDefinition(TEST_PROBLEM_DEF)
    assert dut.metric_configs == {}


OPTIONAL_FIELDS = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "template_disable_list": {},
    "template_enable_list": {},
    "model_disable_list": {},
    "model_enable_list": {},
    "instantiations": [],
    "hyperparam_search_space": {}
}
'''


def test_optional_fields() -> None:
    dut = ProblemDefinition(OPTIONAL_FIELDS)
    assert set(dut._get().keys()).issuperset(set(
        ["template_disable_list", "template_enable_list",
         "model_disable_list", "model_enable_list",
         "instantiations", "hyperparam_search_space"]))


SUNNY_TASK = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "data_type": "text",
        "task": "regression"
    }
}'''


def test_task_sunny_day() -> None:
    dut = ProblemDefinition(SUNNY_TASK)
    prob_type = dut.task
    print(f'Problem type: {prob_type}')
    assert prob_type == Task(clause={'data_type': 'text', 'task': 'regression'})


TASK_WITH_NO_TASK_TYPE = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "data_type": "TABULAR"
    }
}'''


def test_problem_type_missing_task() -> None:
    with pytest.raises(ValidationErrors, match=r'[Tt]ask'):
        ProblemDefinition(TASK_WITH_NO_TASK_TYPE)


TASK_WITH_NO_DATA_TYPE = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "REGRESSION"
    }
}'''


def test_problem_type_missing_data_type() -> None:
    dut = ProblemDefinition(TASK_WITH_NO_DATA_TYPE)
    want = Task(clause={'data_type': "tabular", 'task': "regression"})
    got = dut.task
    assert want == got


NON_DICT_PROBLEM_TYPE = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": []
}'''


def test_problem_type_not_dict() -> None:
    with pytest.raises(ValidationErrors, match=r'dict((.|\n)*)problem_type.task'):
        ProblemDefinition(NON_DICT_PROBLEM_TYPE)


_DEFAULT_INSTANTIATIONS = sorted(['simple', 'stub_executor_kind'])


def test_instantiations_missing() -> None:
    dut = ProblemDefinition(TEST_PROBLEM_DEF)
    instantiations = dut.instantiations
    assert sorted(instantiations) == _DEFAULT_INSTANTIATIONS


EMPTY_INSTANTIATIONS = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "output": {
        "instantiations": []
    }
}'''


def test_instantiations_disable() -> None:
    dut = ProblemDefinition(EMPTY_INSTANTIATIONS)
    instantiations = dut.instantiations
    assert instantiations == []


SIMPLE_INSTANTIATIONS = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "output": {
        "instantiations": [
            "SIMPLE",
            "stub_executor_kind"
        ]
    }
}'''


def test_instantiations_sunny_day() -> None:
    dut = ProblemDefinition(SIMPLE_INSTANTIATIONS)
    instantiations = dut.instantiations
    assert set(instantiations) == set({'simple', 'stub_executor_kind'})


COMMENTS_PRESENT = '''{
    "metrics" : {},
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "_comments": ["Some comment."],
        "data_type": "TEXT",
        "task": "REGRESSION"
    }
}'''


def test_comments() -> None:
    dut = ProblemDefinition(COMMENTS_PRESENT)
    got = dut._get('problem_type')
    want = {
        'data_type': 'TEXT',
        'task': 'REGRESSION'
    }
    assert got == want


BINARY_CLASSIFICATION_NO_POS_LABEL = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": "class"
        }
    },
    "problem_type": {
        "task": "binary_classification"
    },
    "metrics" : [
        {
          "MEAN_ABSOLUTE_ERROR": {}
        }
    ]
}
'''


@pytest.mark.xfail
def test_binary_classification_no_pos_label() -> None:
    with pytest.raises(ValidationErrors, match='binary_classification.*pos_label'):
        ProblemDefinition(BINARY_CLASSIFICATION_NO_POS_LABEL)


BINARY_CLASSIFICATION_WITH_POS_LABEL = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": {
                "name": "class",
                "pos_label": "good"
            }
        }
    },
    "problem_type": {
        "task": "binary_classification"
    }
}
'''


def test_metadata_binary_classification_with_pos_label() -> None:
    dut = ProblemDefinition(BINARY_CLASSIFICATION_WITH_POS_LABEL)
    data = dut.dataset_config.load_train()
    assert data.roles[RoleName.TARGET] == [Column('class')]
    assert data.metadata.pos_labels[RoleName.TARGET] == 'good'


OUTPUT_DIR = '''{{
    "dataset": {{
        "config": "ignore",
        "column_roles": {{}}
    }},
    "problem_type": {{
        "task": "regression"
    }},
    "output": {{
        "path": "{output_path}"
    }}
}}
'''


def test_output_dir_exists(tmp_path: Path) -> None:
    tmp_dir = tmp_path / "sub"
    tmp_dir.mkdir()
    pdef_json = OUTPUT_DIR.format(output_path=str(tmp_dir))
    dut = ProblemDefinition(pdef_json)
    assert dut.output.path == tmp_dir


def test_output_dir_not_exist(tmp_path: Path) -> None:
    newsubdir = tmp_path / "newsubdir"
    pdef_json = OUTPUT_DIR.format(output_path=str(newsubdir))
    dut = ProblemDefinition(pdef_json)
    assert dut.output.path == newsubdir
    assert newsubdir.exists()


def test_output_dir_not_a_dir(tmp_path: Path) -> None:
    notadir = tmp_path / "notadir.txt"
    notadir.write_text('This is a text file, not a directory.\n')
    pdef_json = OUTPUT_DIR.format(output_path=str(notadir))
    with pytest.raises(OutputConfigError):
        _ = ProblemDefinition(pdef_json)


def test_output_dir_parent_not_exist(tmp_path: Path) -> None:
    nonexistent_parent = tmp_path / 'nonexistent' / 'also_nonexistent'
    pdef_json = OUTPUT_DIR.format(output_path=str(nonexistent_parent))
    with pytest.raises(OutputConfigError):
        _ = ProblemDefinition(pdef_json)


FORECASTING_NO_TIME = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": {
                "name": "class"
            }
        },
        "forecasting": {
            "horizon": 5,
            "input_size": 15
        }
    },
    "problem_type": {
        "task": "forecasting"
    }
}
'''


def test_forecasting_no_time() -> None:
    with pytest.raises(ValidationErrors,
                       match='(forecasting.*time)|(time.*forecasting)'):
        ProblemDefinition(FORECASTING_NO_TIME)


FORECASTING_NO_HORIZON = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": {
                "name": "class"
            },
            "time": {
                "name": "FOO"
            }
        },
        "forecasting": {
            "input_size": 15
        }
    },
    "problem_type": {
        "task": "forecasting"
    }
}
'''


def test_forecasting_no_horizon() -> None:
    with pytest.raises(ValidationErrors,
                       match='(forecasting.*horizon)|(horizon.*forecasting)'):
        ProblemDefinition(FORECASTING_NO_HORIZON)


FORECASTING_INVALID_FREQUENCY = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": {
                "name": "class"
            },
            "time": {
                "name": "FOO"
            }
        },
        "forecasting": {
            "horizon": 5,
            "input_size": 15,
            "frequency": "INVALID_FREQ_FOR_TEST"
        }
    },
    "problem_type": {
        "task": "forecasting"
    }
}
'''


def test_forecasting_invalid_frequency() -> None:
    with pytest.raises(ValidationErrors,
                       match='(forecasting.*frequency)|(frequency.*forecasting)'):
        ProblemDefinition(FORECASTING_INVALID_FREQUENCY)


FORECASTING_WITH_FREQUENCY = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": {
                "name": "class"
            },
            "time": {
                "name": "FOO"
            }
        },
        "forecasting": {
            "horizon": 5,
            "input_size": 15,
            "frequency": "D"
        }
    },
    "problem_type": {
        "task": "forecasting"
    }
}
'''


def test_forecasting_with_frequency() -> None:
    dut = ProblemDefinition(FORECASTING_WITH_FREQUENCY)
    got = dut._get('dataset', 'forecasting', 'frequency')
    want = 'D'

    assert got == want


FORECASTING_WITH_DEFAULT_FREQUENCY = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": {
                "name": "class"
            },
            "time": {
                "name": "FOO"
            }
        },
        "forecasting": {
            "horizon": 5,
            "input_size": 15
        }
    },
    "problem_type": {
        "task": "forecasting"
    }
}
'''


def test_forecasting_with_default_frequency() -> None:
    dut = ProblemDefinition(FORECASTING_WITH_DEFAULT_FREQUENCY)
    assert dut.dataset_config.forecasting is not None
    got = dut.dataset_config.forecasting['frequency']
    want = 'M'

    assert got == want


MEMORY_PROBLEM_DEF = '''{
    "dataset": {
        "config": "memory",
        "column_roles": {
            "target": {
                "name": "c"
            }
        }
    },
    "problem_type": {
        "task": "binary_classification"
    }
}
'''


def test_memory_problem_def() -> None:
    dataframe = pd.DataFrame(
        {
            'a': range(1, 15001),
            'b': range(1, 15001),
            'c': range(1, 15001)
        }
    )
    dut = ProblemDefinition(MEMORY_PROBLEM_DEF, train_df=dataframe, test_df=dataframe)
    dataset = dut.dataset_config.load_test()

    assert isinstance(dut.dataset_config, MemoryDatasetConfig)
    pd.testing.assert_frame_equal(dataframe, dut.dataset_config.load_train().dataframe)

    assert dataset is not None
    # drop the target column in the original dataframe because we expect
    #   `load_test()` to drop the target column
    pd.testing.assert_frame_equal(dataframe.drop('c', axis=1), dataset.dataframe)


def test_missing_dataframe_memory_problem_def() -> None:
    # expect error message in the exception has one of the following two patterns:
    #   "memory ... dataframe"
    #   "dataframe ... memory"
    with pytest.raises(ValidationErrors, match=r'(memory.*dataframe)|(dataframe.*memory)/i'):
        ProblemDefinition(MEMORY_PROBLEM_DEF)


MEMORY_PROBLEM_DEF_WITH_K = '''{
    "dataset": {
        "config": "memory",
        "column_roles": {
            "target": {
                "name": "c"
            }
        }
    },
    "problem_type": {
        "task": "binary_classification"
    },
    "cross_validation": {
        "k": 10
    }
}
'''


def test_memory_problem_def_with_k() -> None:
    dataframe = pd.DataFrame(
        {
            'a': range(1, 15001),
            'b': range(1, 15001),
            'c': range(1, 15001)
        }
    )
    dut = ProblemDefinition(MEMORY_PROBLEM_DEF_WITH_K, train_df=dataframe, test_df=dataframe)

    assert dut.cross_validation_config.k == 10


# TODO(piggy): This test relies on validation that was removed to make image processing work.
# The fix is to add a new catalog for problem defs.
@pytest.mark.xfail
def test_memory_problem_def_with_invalid_k() -> None:
    dataframe = pd.DataFrame(
        {
            'a': range(1, 3),
            'b': range(1, 3),
            'c': range(1, 3)
        }
    )

    with pytest.raises(ValidationErrors, match=r'(k.*row)|(row.*k)/i'):
        ProblemDefinition(MEMORY_PROBLEM_DEF_WITH_K, train_df=dataframe, test_df=dataframe)



'''Catalog and autoloader for dataset input formats'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

from .dataset_config import DatasetConfig
from ...catalog.catalog_element_mixin import CatalogElementMixin
from ...catalog.memory_catalog import MemoryCatalog
from ...catalog.plugin_catalog import PluginCatalog


class DatasetFormat(CatalogElementMixin, metaclass=abc.ABCMeta):
    '''Represents a way of reading data into the system.'''
    _builder: Type[DatasetConfig]

    def build(self, clause: Dict[str, Any],
              parents: Optional[List[str]] = None,
              **kwargs) -> DatasetConfig:
        '''Build a dataset config using this class's data format.'''
        return self._builder(clause=clause, parents=parents, **kwargs)


class DatasetFormatCatalog(MemoryCatalog[DatasetFormat], metaclass=abc.ABCMeta):
    '''Base class for dataset config catalogs'''

    def build(self, clause: Dict[str, Any], **kwargs) -> DatasetConfig:
        '''Resolve a dataset config clause to a DatasetConfig.

        This does a catalog lookup for the name in the clause. The lookup
        defaults to 'local'.
        '''
        config = clause.get('config', 'local')
        dataset_format: DatasetFormat = self.lookup_by_name(config)
        return dataset_format.build(clause=clause, **kwargs)


class DatasetFormatCatalogAuto(PluginCatalog[DatasetFormat], DatasetFormatCatalog):
    '''Dataset format catalog that automatically loads all dataset formats from a directory.

    Formats are located in ../dataset_formats.
    '''

    def __init__(self, **kwargs):
        super().__init__(
            catalog_name='dataset_formats',
            default_root=Path(__file__).parents[2], **kwargs)



""" Tests for MemoryDatasetConfig """
from typing import Dict, Any

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd

from .memory import MemoryDatasetConfig
from ..problem_def.task import Task, TaskType
from ..wrangler.dataset import RoleName

# pylint: disable=missing-function-docstring

CLAUSE = {
    'config': 'memory',
    'column_roles': {
        'target': {
            'name': 'a',
            'pos_label': 1
        }
    }
}
TRAIN_DF = pd.DataFrame({'a': [1, 2], 'b': [3, 4]})


def test_sunny_day():
    dut = MemoryDatasetConfig(train_df=TRAIN_DF, clause=CLAUSE)
    dut.validate()
    result = dut.load_train()
    got_train_df = result.dataframe

    assert got_train_df.shape == (2, 2)
    assert RoleName.TARGET in result.metadata.roles.keys()
    assert len(result.metadata.roles[RoleName.TARGET]) == 1
    assert result.metadata.roles[RoleName.TARGET][0].name == 'a'


def test_attribute_role_default():
    dut = MemoryDatasetConfig(train_df=TRAIN_DF, clause=CLAUSE)
    dut.validate()
    result = dut.cols_with_role(role=RoleName.ATTRIBUTE)

    assert 1 == len(result)
    assert result[0].name == 'b'


def test_no_target():
    clause_without_target = {
        'config': 'memory',
        'column_roles': {}
    }
    dut = MemoryDatasetConfig(train_df=TRAIN_DF, clause=clause_without_target)
    dut.validate()

    assert dut.target is None


def test_load_testdata() -> None:
    test_df = pd.DataFrame({'a': [1, 2, 2], 'b': [30, 40, 500]})
    dut = MemoryDatasetConfig(train_df=TRAIN_DF, clause=CLAUSE, test_df=test_df)
    dut.validate()
    result = dut.load_test()

    assert result is not None   # needed to make pylint happy
    got_test_df = result.dataframe

    assert got_test_df.shape == (3, 1)
    assert 'a' not in got_test_df.columns
    assert 'b' in got_test_df.columns


def test_load_nonexistent_testdata() -> None:
    dut = MemoryDatasetConfig(train_df=TRAIN_DF, clause=CLAUSE)
    dut.validate()
    result = dut.load_test()

    assert result is None


FORECASTING_CLAUSE: Dict[str, Any] = {
    'config': 'memory',
    'column_roles': {
        'target': {
            'name': 'a'
        }
    },
    'forecasting': {
        'horizon': 1,
        'input_size': 1,
        'frequency': 'M'
    }
}


FORECASTING_TASK_CLAUSE = {
    'task': 'forecasting'
}


def test_load_testdata_forecasting() -> None:
    """check that we don't drop the target column for forecasting test data"""
    forecasting_clause = FORECASTING_CLAUSE.copy()
    test_df = pd.DataFrame({'a': [1], 'b': [30]})

    dut = MemoryDatasetConfig(
        train_df=TRAIN_DF,
        clause=forecasting_clause,
        test_df=test_df,
        task=Task(clause=FORECASTING_TASK_CLAUSE))
    dut.validate()
    test_data = dut.load_test()

    assert test_data is not None
    assert 'a' in test_data.dataframe.columns


def test_forecasting_metadata() -> None:
    forecasting_clause = {
        'config': 'memory',
        'column_roles': {
            'target': {
                'name': 'a'
            }
        },
        'forecasting': {
            'horizon': 1,
            'input_size': 3,
            'frequency': 'M'
        }
    }
    custom_train_df = pd.DataFrame({'a': [1, 2, 1, 2, 1], 'b': [3, 4, 5, 6, 7]})

    dut = MemoryDatasetConfig(
        train_df=custom_train_df,
        clause=forecasting_clause,
        task=Task(clause=FORECASTING_TASK_CLAUSE))
    dut.validate()
    train_data = dut.load_train()

    assert train_data.metadata.horizon == 1
    assert train_data.metadata.input_size == 3
    assert train_data.metadata.frequency == 'M'
    assert train_data.metadata.step_size == 1


CLASSIFICATION_TASK_CLAUSE = {
    'task': 'binary_classification'
}


def test_task_in_metadata_with_load_train() -> None:
    clause = CLAUSE.copy()
    task = Task(clause=CLASSIFICATION_TASK_CLAUSE)
    dut = MemoryDatasetConfig(train_df=TRAIN_DF, clause=clause, task=task)
    dut.validate()
    result = dut.load_train()

    assert result.metadata.task == TaskType.BINARY_CLASSIFICATION


def test_task_in_metadata_with_load_test() -> None:
    clause = CLAUSE.copy()
    test_df = pd.DataFrame({'a': [1], 'b': [30]})
    task = Task(clause=CLASSIFICATION_TASK_CLAUSE)
    dut = MemoryDatasetConfig(train_df=TRAIN_DF, test_df=test_df,
                              clause=clause, task=task)
    dut.validate()
    result = dut.load_test()

    assert result is not None
    assert result.metadata.task == TaskType.BINARY_CLASSIFICATION



''' Base class for dataset configs that use dataframes '''
import abc
from typing import Any, Optional, List, Union

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd

from ...problem_def.config_component import ConfigError, ValidationErrors
from .dataset_config import DatasetConfig, DatasetConfigError, ParsingErrors, DatasetFileError
from ...problem_def.task import TaskType
from ...wrangler.dataset import Dataset, Column, RoleName, DatasetKeys


class DataframeDatasetConfig(DatasetConfig, metaclass=abc.ABCMeta):
    '''Base class for dataset configs that use dataframes
    '''
    _train_cols: pd.DataFrame

    @abc.abstractmethod
    def load_train(self) -> Dataset:
        ...

    @abc.abstractmethod
    def load_test(self) -> Optional[Dataset]:
        ...

    def dataset(self,
                data: Any,
                key: Union[str, DatasetKeys] = DatasetKeys.DATAFRAME,
                cols: Optional[List[str]] = None,
                roles: Optional[List[Union[RoleName, str]]] = None,
                **kwargs) -> Dataset:
        '''Load a Dataset object, by placing data at the supplied key.

        As of 2023-12-04 we only know how to handle things we can turn into a pandas DataFrame.

        Args:
          data:
            Dataframe or object that can be turned into one.
          key:
            The key for the data in the dataset. Defaults to "dataframe".
          cols:
            If set, only selects supplied column(s).
            Will take union with roles if both are specified.
          roles:
            If set, only selects columns with supplied role(s).
            Will take union with cols if both are specified.
        '''

        retval = Dataset(metadata=self.metadata)

        data_df = pd.DataFrame(data)

        # Trim to selected columns if either cols or roles is set.
        cols_to_select = set()
        if cols is not None:
            cols_to_select.update(cols)
        if roles is not None:
            for role in roles:
                cols_to_select.update(
                    c.name for c in self.cols_with_role(role))
        if cols_to_select:
            data_df = data_df[sorted(cols_to_select)]

        if isinstance(key, str):
            key = DatasetKeys[key.upper()]

        retval[key.value] = data_df
        return retval

    def validate(self, **kwargs) -> None:
        '''Look for errors by comparing the problem definition JSON to the
            dataframe that was read.'''
        super().validate(**kwargs)
        errors: List[ConfigError] = []
        # TODO(Merritt): add test case for this validation error
        for rolename, cols in self._roles.items():
            for col in cols:
                if col.name not in self._train_cols.columns:
                    errors.append(DatasetFileError(
                        f'Column {col.name} specified under role {rolename}'
                        ' not found among columns in dataset. '
                        f'Found: {self._train_cols.columns}'))
        if len(errors) != 0:
            raise ValidationErrors(errors)

    def _read_roles(self) -> None:
        '''Flesh out roles based on the actual dataset.

        Give all unassigned columns the Attribute role.
        '''
        assigned_colnames: List[str] = []
        errors: List[DatasetConfigError] = []
        for cols in self.roles.values():
            for col in cols:
                assert col.name is not None
                assigned_colnames.append(str(col.name))

        unassigned_colnames = list(set(self._train_cols.columns) - set(assigned_colnames))
        unassigned_cols = []
        for colname in unassigned_colnames:
            unassigned_cols.append(Column(colname))

        # TODO(Merritt): Deal with situation where the user manually specifies a col as attribute
        self._roles[RoleName.ATTRIBUTE] = unassigned_cols

        if len(errors) != 0:
            raise ParsingErrors(errors)

    def _prune_target(self, retval: Dataset) -> Dataset:
        '''
        remove target col if it is in data; test data shouldn't have target
        keep target for forecasting problems
        '''

        if self._task is None or self._task.task_type != TaskType.FORECASTING:
            if self.target is not None and self.target.name in retval.dataframe.columns:
                new_data = retval.dataframe.drop([self.target.name], axis=1)
                retval.dataframe = new_data

        return retval



'''Output configuration'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path
from typing import Optional, Set, Dict, Any, List

from .config_component import ConfigComponent, ConfigError
from ..wrangler.constants import FileType, Defaults, ProblemDefKeys
from ..executor.executor_kind import ExecutorKind


class OutputConfigError(ConfigError):
    '''Errors during building or validating OutputConfig.'''


class OutputConfig(ConfigComponent):
    '''Output configuration'''

    def __init__(self, clause: Dict[str, Any], parents: Optional[List[str]] = None):
        parents = self._add_parent(parents, ProblemDefKeys.OUTPUT.value)
        super().__init__(clause=clause, parents=parents)

    def validate(self, **kwargs) -> None:
        if self.path is not None:
            if self.path.exists() and not self.path.is_dir():
                raise OutputConfigError(
                    f'Output path {self.path} is not a directory'
                )

            if not self.path.parent.exists():
                raise OutputConfigError(f'Parent of path {self.path} does not exist.')

            if not self.path.parent.is_dir():
                raise OutputConfigError(f'Parent of path {self.path} is not a directory.')

            # Confirm that we have the needed permissions to create this directory.
            self.path.mkdir(exist_ok=True)

        self.file_type  # pylint: disable=pointless-statement

        try:
            # will raise an error if any instantiations are not allowed
            _ = self.instantiations
        except NotImplementedError as err:
            raise OutputConfigError(err) from err

    @property
    def path(self) -> Optional[Path]:
        '''Path to the output folder specified in the problem def'''
        if self._exists(ProblemDefKeys.OUTPUT_PATH):
            return Path(self._get(ProblemDefKeys.OUTPUT_PATH)).resolve()
        return None

    @property
    def instantiations(self) -> Set[ExecutorKind]:
        '''Kinds of instantiations specified in the problem'''
        retval = self._get_with_default(
            ProblemDefKeys.INSTATIATIONS, dflt=Defaults.INSTANTIANTIONS)
        return set(ExecutorKind(s) for s in retval)

    @property
    def file_type(self) -> FileType:
        '''Type of file for predictions to be saved in.'''
        retval = self._get_with_default(ProblemDefKeys.FILE_TYPE.value, dflt=Defaults.FILE_TYPE)
        try:
            return FileType[retval.upper()]
        except KeyError as exc:
            raise OutputConfigError(f'{retval} is not a valid file type') from exc



'''Tests for LocalDatasetConfig'''
from pathlib import Path
from typing import Any, Dict, Optional
import os

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd
import pytest

from ..problem_def.problem_def import ValidationErrors
from ..problem_def.task import Task, TaskType
from ..wrangler.dataset import DatasetKeys, RoleName
from .local import LocalDatasetConfig


# pylint: disable=missing-function-docstring


def valid_csv(filename: Optional[str] = None) -> str:
    '''Returns a path (in the form of a string) to a valid csv file.'''
    current_path = str(os.getenv('PYTEST_CURRENT_TEST')).split('::', maxsplit=1)[0]
    pathobj = Path(current_path).resolve()
    module_parent = pathobj.parents[2]
    if filename is None:
        filename = 'credit.csv'
    path = module_parent / 'examples' / 'classification' / filename
    return str(path)


def valid_arff(filename: Optional[str] = None) -> str:
    '''Returns a path (in the form of a string) to a valid arff file.'''
    current_path = str(os.getenv('PYTEST_CURRENT_TEST')).split('::', maxsplit=1)[0]
    pathobj = Path(current_path).resolve()
    module_parent = pathobj.parents[2]
    if filename is None:
        filename = 'dataset_31_credit-g.arff'
    path = module_parent / 'examples' / 'classification' / filename
    return str(path)


def make_dataset(train_path: str) -> Dict[str, Any]:
    retval = {
        'config': 'local',
        'train_path': train_path,
        'column_roles': {
            'target': {
                'name': 'class',
                'pos_label': 'good'
            }
        }
    }
    return retval


def test_load_simple_dataset():
    clause = make_dataset(valid_csv())
    dut = LocalDatasetConfig(clause=clause)
    result = dut.load_train()

    train_df = result.dataframe
    assert train_df.shape == (1000, 21)
    assert train_df.columns[0] == 'checking_status'
    assert train_df.columns[20] == 'class'


def test_cols_with_role():
    dut = LocalDatasetConfig(clause=make_dataset(valid_csv()))
    result = dut.cols_with_role(role=RoleName.TARGET)
    assert 1 == len(result)
    assert 'class' == result[0].name


def test_attribute_role_default():
    dut = LocalDatasetConfig(clause=make_dataset(valid_csv()))
    result = dut.cols_with_role(role=RoleName.ATTRIBUTE)
    assert 20 == len(result)


def test_no_target():
    clause = {
        'config': 'local',
        'train_path': valid_csv(),
        'column_roles': {
            'index': {
                'id': 0
            }
        }
    }
    dut = LocalDatasetConfig(clause=clause)
    assert dut.target is None


def test_arff():
    clause = {
        'config': 'local',
        'train_path': valid_arff(),
        'column_roles': {
            'target': {
                'name': 'class'
            }
        }
    }
    dut = LocalDatasetConfig(clause=clause)
    assert isinstance(dut, LocalDatasetConfig)
    assert 'class' == dut.target.name
    assert len(dut.index_cols) == 0


def test_load_arff():
    clause = {
        'config': 'local',
        'train_path': valid_arff(),
        'column_roles': {
            'target': {
                'name': 'class'
            }
        }
    }
    dut = LocalDatasetConfig(clause=clause)
    result = dut.load_train()

    assert DatasetKeys.DATAFRAME.value in result

    train_df = result.dataframe
    assert isinstance(train_df, pd.DataFrame)
    assert train_df.shape == (1000, 21)


def test_load_testdata() -> None:
    clause = {
        'config': 'simple',
        'train_path': valid_arff(filename='credit-train.csv'),
        'test_path': valid_arff(filename='credit-test.csv'),
        'column_roles': {
            'target': {
                'name': 'class'
            }
        }
    }
    dut = LocalDatasetConfig(clause=clause)
    result = dut.load_test()
    assert result is not None
    test_df = result.dataframe
    assert test_df.shape == (200, 20)
    assert 'class' not in test_df.columns
    assert 'own_telephone' in test_df.columns


def test_load_nonexistent_testdata() -> None:
    '''If there is no test data, load_test() returns None'''
    clause = make_dataset(valid_csv())
    dut = LocalDatasetConfig(clause=clause)
    dut.validate()
    result = dut.load_test()
    assert result is None


FORECASTING_CLAUSE: Dict[str, Any] = {
    'config': 'local',
    'column_roles': {
        'target': {
            'name': 'class'
        }
    },
    'forecasting': {
        'horizon': 5,
        'input_size': 15,
        'frequency': 'M'
    }
}


FORECASTING_TASK_CLAUSE = {
    'task': 'forecasting'
}


def test_load_testdata_forecasting() -> None:
    # check that we don't drop the target column for forecasting test data
    clause = FORECASTING_CLAUSE.copy()
    clause['train_path'] = valid_csv()
    clause['test_path'] = valid_csv()

    dut = LocalDatasetConfig(
        clause=clause,
        task=Task(clause=FORECASTING_TASK_CLAUSE))
    test_data = dut.load_test()
    assert test_data is not None
    assert 'class' in test_data.dataframe.columns


def test_forecasting_metadata() -> None:
    clause = FORECASTING_CLAUSE.copy()
    clause['train_path'] = valid_csv()

    dut = LocalDatasetConfig(
        clause=clause,
        task=Task(clause=FORECASTING_TASK_CLAUSE))
    dut.validate()
    assert dut.load_train().metadata.horizon == 5
    assert dut.load_train().metadata.input_size == 15
    assert dut.load_train().metadata.frequency == 'M'
    assert dut.load_train().metadata.step_size == 5


def test_forecasting_step_size() -> None:
    clause = FORECASTING_CLAUSE.copy()
    clause['train_path'] = valid_csv()
    forecasting_dict = clause['forecasting']
    new_dict = forecasting_dict.copy()
    new_dict['step_size'] = 3
    clause['forecasting'] = new_dict

    dut = LocalDatasetConfig(
        clause=clause,
        task=Task(clause=FORECASTING_TASK_CLAUSE))
    dut.validate()
    assert dut.load_train().metadata.horizon == 5
    assert dut.load_train().metadata.input_size == 15
    assert dut.load_train().metadata.frequency == 'M'
    assert dut.load_train().metadata.step_size == 3


def test_forecasting_invalid_step_size() -> None:
    clause = FORECASTING_CLAUSE.copy()
    clause['train_path'] = valid_csv()
    forecasting_dict = clause['forecasting']
    new_dict = forecasting_dict.copy()
    new_dict['step_size'] = 8
    clause['forecasting'] = new_dict

    dut = LocalDatasetConfig(
        clause=clause,
        task=Task(clause=FORECASTING_TASK_CLAUSE))

    with pytest.raises(ValidationErrors, match=r"step_size"):
        dut.validate()


def test_static_exog() -> None:
    clause = FORECASTING_CLAUSE.copy()
    clause['train_path'] = valid_csv()
    clause['test_path'] = valid_csv()
    clause['static_exogenous_path'] = valid_csv()
    dut = LocalDatasetConfig(
        clause=clause,
        task=Task(clause=FORECASTING_TASK_CLAUSE))
    dut.validate()
    got = dut.load_train()
    assert isinstance(got['static_exogenous'], pd.DataFrame)
    got_test = dut.load_test()
    assert got_test is not None
    assert isinstance(got_test['static_exogenous'], pd.DataFrame)
    pd.testing.assert_frame_equal(got['static_exogenous'], got_test['static_exogenous'])


def test_static_exog_missing() -> None:
    clause = FORECASTING_CLAUSE.copy()
    clause['train_path'] = valid_csv()
    clause['test_path'] = valid_csv()

    dut = LocalDatasetConfig(
        clause=clause,
        task=Task(clause=FORECASTING_TASK_CLAUSE))
    dut.validate()
    got = dut.load_train()
    assert got['static_exogenous'] is None

    got_test = dut.load_test()
    assert got_test is not None
    assert got_test['static_exogenous'] is None


CLASSIFICATION_TASK_CLAUSE = {
    'task': 'binary_classification'
}


def test_task_in_metadata_with_load_train() -> None:
    clause = make_dataset(valid_csv())
    task = Task(clause=CLASSIFICATION_TASK_CLAUSE)
    dut = LocalDatasetConfig(clause=clause, task=task)
    dut.validate()
    result = dut.load_train()

    assert result.metadata.task == TaskType.BINARY_CLASSIFICATION


def test_task_in_metadata_with_load_test() -> None:
    clause = {
        'config': 'simple',
        'train_path': valid_arff(filename='credit-train.csv'),
        'test_path': valid_arff(filename='credit-test.csv'),
        'column_roles': {
            'target': {
                'name': 'class'
            }
        }
    }
    task = Task(clause=CLASSIFICATION_TASK_CLAUSE)
    dut = LocalDatasetConfig(clause=clause, task=task)
    dut.validate()
    result = dut.load_test()

    assert result is not None
    assert result.metadata.task == TaskType.BINARY_CLASSIFICATION


def test_load_arbitrary_csv() -> None:
    dut = LocalDatasetConfig(make_dataset(valid_csv(filename='credit-train.csv')))
    got = dut.dataset(data=valid_csv(filename='credit-test.csv'),
                      cols=['class', 'own_telephone'])
    assert got.dataframe.shape == (200, 2)
    assert set(got.dataframe.columns) == {'class', 'own_telephone'}
    assert got.metadata.roles[RoleName.TARGET][0].name == 'class'



'''Class representing problem type, which is a combination of data type and task.'''
from typing import List, Optional, Dict, Any
from enum import Enum, auto

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .config_component import ConfigComponent, ConfigError
from .config_component import InvalidKeyError, ValidationErrors, MissingKeyError
from .config_component import InvalidValueError, ProblemDefTypeError
from ..wrangler.constants import ProblemDefKeys, ProblemDefKeySet, Defaults


class DataType(Enum):
    '''Possible input data types'''
    TABULAR = auto()
    IMAGE = auto()
    TIMESERIES = auto()
    TEXT = auto()

    @staticmethod
    def validate(name: str) -> List[str]:
        '''Return an error message if data type is not in the list'''
        errors: List[str] = []
        names = [member.name for member in DataType]
        if name.upper() not in names:
            errors.append(f'"{name}" not among valid data types: f{names}')
        return errors


class TaskType(Enum):
    '''Possible ML tasks'''
    BINARY_CLASSIFICATION = auto()
    MULTICLASS_CLASSIFICATION = auto()
    REGRESSION = auto()
    FORECASTING = auto()
    TEST_TASK = auto()  # Only used in testing

    @staticmethod
    def validate(name: str) -> List[str]:
        '''Return an error message if task type is not in the list'''
        errors: List[str] = []
        names = [member.name for member in TaskType]
        if name.upper() not in names:
            errors.append(f'"{name}" not among valid task types: f{names}')
        return errors


class Task(ConfigComponent):
    '''Parsed form of the problem_type.'''

    def __init__(self, clause: Dict[str, Any], parents: Optional[List[str]] = None):
        parents = self._add_parent(parents, ProblemDefKeys.TASK.value)
        super().__init__(clause=clause, parents=parents)

    @property
    def data_type(self) -> DataType:
        '''Data type'''
        return DataType[self._get_with_default(
            ProblemDefKeys.DATA_TYPE.value,
            dflt=Defaults.DATA_TYPE).upper()]

    @property
    def task_type(self) -> Optional[TaskType]:
        '''ML task'''
        field = self._get_with_default(
            ProblemDefKeys.TASK_TYPE.value,
            dflt=None
        )
        if field is None:
            return None
        try:
            return TaskType[field.upper()]
        except KeyError as err:
            raise InvalidValueError(
                f'task type {field} is not recognized.') from err

    def __str__(self) -> str:
        data_type_name = "None"
        if self.data_type is not None:
            data_type_name = self.data_type.name
        task_name = "None"
        if self.task_type is not None:
            task_name = self.task_type.name
        return (f'{{Data type: {data_type_name}'
                f', Task: {task_name} }}')

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, Task):
            if __value.data_type == self.data_type and __value.task_type == self.task_type:
                return True
        return False

    def validate(self, **kwargs) -> None:
        # test that problem_type maps to a dictionary containing exactly data_type and task_type
        errors: List[ConfigError] = []

        if isinstance(self._clause, dict):
            task_keys = set(self._clause.keys())
            if not ProblemDefKeySet.TASK.REQUIRED.issubset(task_keys):
                errors.append(MissingKeyError(
                    'Required keys missing in problem type: '
                    f'{ProblemDefKeySet.TASK.REQUIRED.difference(task_keys)}'))

            if not task_keys.issubset(ProblemDefKeySet.TASK.ALLOWED):
                errors.append(InvalidKeyError(
                    'Invalid key(s) in problem type: '
                    f'{task_keys.difference(ProblemDefKeySet.TASK.ALLOWED)}'
                ))

            if ProblemDefKeys.DATA_TYPE.value in self._clause:
                errmessages = DataType.validate(self._clause[ProblemDefKeys.DATA_TYPE.value])
                errors = errors + [InvalidValueError(err) for err in errmessages]

            if ProblemDefKeys.TASK_TYPE.value in self._clause:
                errmessages = TaskType.validate(self._clause[ProblemDefKeys.TASK_TYPE.value])
                errors = errors + [InvalidValueError(err) for err in errmessages]
        else:
            errors.append(ProblemDefTypeError(
                f'{ProblemDefKeys.TASK.value} must be a dict, '
                f'instead found {type(self._clause)}'))

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)



'''Configuration for the hyperparam clause.

 We use a single data structure, HyperparamConfig, for all hyperparam clauses
 instead of separate structures for each clause so that it is easier to do
 intelligent searches of hyperparameter overrides.
'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, Any, List, Union

from ..searcher.params import Overrides, Override, Selector, ParamRanges, ParamRange
from ..wrangler.constants import (ProblemDefControl, ProblemDefKeys,
                                  RangeMethod, Matcher, MATCHERS_TO_LOWERCASE)

from .config_component import ConfigComponent, ConfigError


class HyperparamError(ConfigError):
    '''Accummulating error for hyperparam parsing.'''

    errors: List[ConfigError]

    def __init__(self, errors: List[ConfigError]):
        super().__init__(f'At least one validation error occured: {errors!r}')
        self.errors = errors


class HyperparamTooManyClauses(ConfigError):
    '''A subclause had unexpected keys.'''


class ControlError(ConfigError):
    '''The control strings for hyperparam config are incorrect.'''


class HyperparamConfig(ConfigComponent):
    '''Configuration for the hyperparam clause.

    We use a single data structure, HyperparamConfig, for all hyperparam clauses
    instead of separate structures for each clause so that it is easier to do
    intelligent searches of hyperparameter overrides.
    '''
    _control: List[str]
    _disable_grid_search = False

    def __init__(self, clause: Dict[str, Any]):
        super().__init__(clause=clause)
        self._overrides = Overrides()

        (overrides, control) = self._split_overrides_control(self._get(ProblemDefKeys.HYPERPARAMS))
        for override_clause in overrides:
            self._overrides.append(self._parse_override(override_clause))
        self._control = control
        self._parse_control()

        self.validate()

    def _parse_control(self):
        for s in self._control:
            if s == ProblemDefControl.DISABLE_GRID_SEARCH.value:
                self._disable_grid_search = True
            else:
                raise ControlError(f'Invalid control {s}')

    def _split_overrides_control(self, clauses: List[Union[str, Dict[str, Any]]]):
        retval_str = []
        retval_overrides = []
        for val in clauses:
            if isinstance(val, str):
                retval_str.append(val)
            elif isinstance(val, dict):
                retval_overrides.append(val)
            else:
                raise ControlError(f'Unexpected hyperparam type, {val}:{type(val)}')
        return (retval_overrides, retval_str)

    def _parse_override(self, subclause: Dict[str, Dict[str, Any]]) -> Override:
        select_clause = subclause.pop(ProblemDefKeys.SELECT.value)
        params_clause = subclause.pop(ProblemDefKeys.PARAMS.value)
        if subclause:
            raise HyperparamTooManyClauses(f'Unexpected extra clauses: {subclause!r}')
        select = Selector()
        for matcher, arg in select_clause.items():
            matcher_enum = Matcher[matcher.upper()]
            if matcher_enum in MATCHERS_TO_LOWERCASE:
                arg = arg.lower()
            select[matcher_enum] = arg
        params = ParamRanges()
        for name, hyperparam_param in params_clause.items():
            params[name] = self._parse_param(hyperparam_param)

        return Override(selector=select, params=params)

    def _parse_param(self, param: Dict[str, Any]) -> ParamRange:
        default = param.pop(ProblemDefKeys.DEFAULT.value, None)
        if len(param) != 1:
            raise HyperparamTooManyClauses(f'expecting 1 key, got {param!r}')

        for method, range_value in param.items():
            return ParamRange(method=RangeMethod[method.upper()],
                              prange=range_value,
                              default=default)
        assert False, 'BUG: Unreachable code'

    @property
    def disable_grid_search(self) -> bool:
        """Should we disable grid search?

        If true, we skip the Searcher.
        """
        return self._disable_grid_search

    @property
    def overrides(self) -> Overrides:
        '''Return the hyperparam overrides described by the config.'''
        return self._overrides

    def validate(self, **kwargs) -> None:
        '''Make sure the hyperparam object makes sense.'''
        allowed = {x.value for x in ProblemDefControl}
        if not set(self._control).issubset(allowed):
            not_recognized = set(self._control) - allowed
            raise ControlError(
                f'Unrecognized control options: {not_recognized}'
                f'Must be among {allowed}'
            )



'''Stub dataset config for tests.'''

from typing import Any, Optional

import pandas as pd

from ..wrangler.dataset import Dataset, DatasetKeys
from .impl.dataset_config import DatasetConfig
from .impl.dataset_format_catalog import DatasetFormat, DatasetFormatCatalog


class StubDatasetConfig(DatasetConfig):
    '''Stub dataset used in testing.'''

    def load_train(self) -> Dataset:
        '''Load a stub dataset'''
        return Dataset(metadata=self.metadata, **{DatasetKeys.DATAFRAME.value: pd.DataFrame()})

    def load_test(self) -> Optional[Dataset]:
        '''Load test data if it exists, otherwise None'''
        return None

    def dataset(self, data: Any, **kwargs) -> Dataset:
        return Dataset(metadata=self.metadata, **{DatasetKeys.DATAFRAME.value: pd.DataFrame()})


class StubFormat(DatasetFormat):
    '''Stub dataset format used in testing.'''
    _builder = StubDatasetConfig
    _name = 'ignore'


def register(catalog: DatasetFormatCatalog, *args, **kwargs):
    '''Register all the objects in this file.'''
    catalog.register(StubFormat(*args, **kwargs))



'''Tests for dataset_config_catalog.py'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .dataset_format_catalog import DatasetFormatCatalogAuto
from ..memory import MemoryDatasetFormat
from ..local import LocalDatasetFormat

# pylint: disable=missing-function-docstring


def test_sunny_day() -> None:
    dut = DatasetFormatCatalogAuto()
    assert isinstance(dut.lookup_by_name('memory'), MemoryDatasetFormat)
    assert isinstance(dut.lookup_by_name('local'), LocalDatasetFormat)



'''Holds a dataset.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
from typing import Any, Dict, List, Optional

import pandas as pd

from ..problem_def.task import DataType, TaskType


class Error(BaseException):
    '''Base class for all model-related exceptions.'''


class DatasetKeyError(Error, KeyError):
    '''Raise when a fit() or predict() method is passed an input dataset
    that is missing required keys.'''


class DatasetValueError(Error, ValueError):
    '''Raise when a dataset has the correct keys but invalid value(s)'''


class RoleName(Enum):
    '''Possible column roles'''
    INDEX = 'index'
    TARGET = 'target'
    ATTRIBUTE = 'attribute'

    # For testing only
    TEST_ROLE = 'test_role'

    # For time series only
    TIME = 'time'
    TIMESERIES_ID = 'timeseries_id'
    PAST_EXOGENOUS = 'past_exogenous'
    FUTURE_EXOGENOUS = 'future_exogenous'


class Column():
    '''Defines a column in a dataset'''
    _name: str

    def __init__(self,
                 name: str):
        self._name = name

    @property
    def name(self) -> str:
        '''The column's name'''
        return str(self._name)

    def __eq__(self, other: Any):
        if not isinstance(other, Column):
            return False
        return self.name == other.name

    def __str__(self) -> str:
        return f'Column({self.name})'

    def __repr__(self) -> str:
        return f'ngautonml.wrangler.dataset.Column ({self.name})'


class DatasetKeys(Enum):
    '''Standard keys used in the Dataset object passed between models'''
    DATAFRAME = 'dataframe'
    TARGET = 'target'
    COVARIATES = 'covariates'
    PREDICTIONS = 'predictions'
    PROBABILITIES = 'probabilities'
    TRAIN = 'train'
    VALIDATE = 'validate'
    GROUND_TRUTH = 'ground_truth'
    ERROR = 'error'

    # for time series forecasting only
    STATIC_EXOGENOUS = 'static_exogenous'
    DYNAMIC_EXOGENOUS = 'dynamic_exogenous'

    # for keras image processing only
    KERAS_DS = 'keras_ds'
    KERAS_VALIDATE = 'keras_validate'


class Metadata():
    '''Metadata.'''
    _roles: Dict[RoleName, List[Column]]
    _pos_labels: Dict[RoleName, Any]
    _task: Optional[TaskType]
    _data_type: Optional[DataType]
    # Forecast related metadata
    _horizon: Optional[int]
    _input_size: Optional[int]
    _frequency: Optional[str]
    _step_size: Optional[int]

    def __init__(self,
                 roles: Optional[Dict[RoleName, List[Column]]] = None,
                 pos_labels: Optional[Dict[RoleName, Any]] = None,
                 forecasting: Optional[Dict[str, Any]] = None,
                 task: Optional[TaskType] = None,
                 data_type: Optional[DataType] = None) -> None:
        self._roles = roles or {}
        self._pos_labels = pos_labels or {}
        self._horizon = forecasting.get('horizon') if forecasting else None
        self._input_size = forecasting.get('input_size') if forecasting else None
        self._frequency = forecasting.get('frequency') if forecasting else None
        self._step_size = forecasting.get('step_size') if forecasting else None
        self._task = task
        self._data_type = data_type

    @property
    def roles(self) -> Dict[RoleName, List[Column]]:
        '''roles maps role names to lists of columns with that role.'''
        return self._roles.copy()

    @property
    def task(self) -> Optional[TaskType]:
        '''Data science task that is being approached with this dataset.'''
        return self._task

    @property
    def data_type(self) -> Optional[DataType]:
        '''Type of data contained in this dataset (image, tabular, etc)'''
        return self._data_type

    @property
    def target(self) -> Optional[Column]:
        '''Get name and index of target column if it exists.'''
        if RoleName.TARGET not in self._roles or len(self._roles[RoleName.TARGET]) == 0:
            return None
        assert len(self._roles[RoleName.TARGET]) <= 1, 'Must have at most 1 target column'
        return self._roles[RoleName.TARGET][0]

    @property
    def pos_labels(self) -> Dict[RoleName, Any]:
        '''maps role names to a positive value for columns with that role'''
        return self._pos_labels.copy()

    @property
    def horizon(self) -> Optional[int]:
        '''Gives the forecast horizon (None if not a forecasting problem).'''
        return self._horizon

    @property
    def input_size(self) -> Optional[int]:
        '''Returns the input size of the data (None if not a forecasting problem).'''
        return self._input_size

    @property
    def frequency(self) -> Optional[str]:
        '''Returns the frequency of the data (None if not a forecasting problem).'''
        return self._frequency

    @property
    def step_size(self) -> Optional[int]:
        '''Returns the step_size of the data (None if not a forecasting problem).'''
        return self._step_size

    def override_roles(self, roles=Dict[RoleName, List[Column]]) -> 'Metadata':
        '''Return a copy of metadata with new roles.'''
        return self.__class__(
            roles=roles,
            pos_labels=self._pos_labels,
            forecasting={'horizon': self._horizon,
                         'input_size': self._input_size,
                         'frequency': self._frequency})


class Dataset(Dict[str, Any]):
    '''Holds a dataset.
    Allows for arbitrary keys and values as long as the keys are strings.

    Standard keys: covariates, target, dataframe
    '''
    _metadata: Metadata

    def __init__(self,
                 *args,
                 metadata: Optional[Metadata] = None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        if metadata is None:
            metadata = Metadata(roles={}, pos_labels={})
        self._metadata = metadata

    @property
    def roles(self) -> Dict[RoleName, List[Column]]:
        '''Information about column roles in the dataset'''
        return self._metadata.roles

    @property
    def metadata(self) -> Metadata:
        '''Config metadata for use by models'''
        return self._metadata

    @property
    def dataframe(self) -> pd.DataFrame:
        '''Return the dataframe at DatasetKeys.DATAFRAME, or raise an error if it doesn't exist.'''

        if DatasetKeys.DATAFRAME.value not in self:
            raise DatasetKeyError(
                f'Attempting to extract key "{DatasetKeys.DATAFRAME.value}" '
                f'as input, instead found keys {list(self.keys())}.')
        df_value = self[DatasetKeys.DATAFRAME.value]
        if not isinstance(df_value, pd.DataFrame):
            raise DatasetValueError(
                f'Input key key {DatasetKeys.DATAFRAME.value} must point to '
                f'a pandas.DataFrame, instead found a {type(df_value)}: '
                f'{df_value}')
        return pd.DataFrame(df_value)

    @dataframe.setter
    def dataframe(self, data: pd.DataFrame) -> None:
        self[DatasetKeys.DATAFRAME.value] = data

    @property
    def ground_truth(self) -> pd.DataFrame:
        '''Return the dataframe at DatasetKeys.GROUND_TRUTH

        or raise an error if it doesn't exist.'''

        if DatasetKeys.GROUND_TRUTH.value not in self:
            raise DatasetKeyError(
                f'Attempting to extract key "{DatasetKeys.GROUND_TRUTH.value}" '
                f'as input, instead found keys {list(self.keys())}.')
        df_value = self[DatasetKeys.GROUND_TRUTH.value]
        if not isinstance(df_value, pd.DataFrame):
            raise DatasetValueError(
                f'Input key key {DatasetKeys.GROUND_TRUTH.value} must point to '
                f'a pandas.DataFrame, instead found a {type(df_value)}: '
                f'{df_value}')
        return pd.DataFrame(df_value)

    @ground_truth.setter
    def ground_truth(self, data: pd.DataFrame) -> None:
        self[DatasetKeys.GROUND_TRUTH.value] = data

    @property
    def predictions(self) -> pd.DataFrame:
        '''Return the dataframe at DatasetKeys.PREDICTIONS
        or raise an error if it doesn't exist.'''

        if DatasetKeys.PREDICTIONS.value not in self:
            raise DatasetKeyError(
                f'Attempting to extract key "{DatasetKeys.PREDICTIONS.value}" '
                f'as input, instead found keys {list(self.keys())}.')
        df_value = self[DatasetKeys.PREDICTIONS.value]
        if not isinstance(df_value, pd.DataFrame):
            raise DatasetValueError(
                f'Input key key {DatasetKeys.PREDICTIONS.value} must point to '
                f'a pandas.DataFrame, instead found a {type(df_value)}: '
                f'{df_value}')
        return pd.DataFrame(df_value)

    @predictions.setter
    def predictions(self, data: pd.DataFrame) -> None:
        self[DatasetKeys.PREDICTIONS.value] = data

    @property
    def probabilities(self) -> pd.DataFrame:
        '''
        Return the dataframe at `DatasetKeys.PROBABILITIES`
        or raise an error if it doesn't exist.
        '''

        if DatasetKeys.PROBABILITIES.value not in self:
            raise DatasetKeyError(
                f'Attempting to extract key "{DatasetKeys.PROBABILITIES.value}" '
                f'as input, instead found keys {list(self.keys())}.'
            )

        df_value = self[DatasetKeys.PROBABILITIES.value]

        if not isinstance(df_value, pd.DataFrame):
            raise DatasetValueError(
                f'Input key key {DatasetKeys.PROBABILITIES.value} must point to '
                f'a pandas.DataFrame, instead found a {type(df_value)}: '
                f'{df_value}'
            )

        return pd.DataFrame(df_value)

    @probabilities.setter
    def probabilities(self, data: pd.DataFrame) -> None:
        self[DatasetKeys.PROBABILITIES.value] = data

    def output(self, override_metadata: Optional[Metadata] = None) -> 'Dataset':
        '''Return a new dataset containing same metadata and nothing else.

        Metadata can also be overriden, all preexisting metadata will be lost.
        Generally called to create a new dataset to fill with a model's output.
        '''
        return self.__class__(metadata=override_metadata or self._metadata)

    def get_dataframe(self) -> pd.DataFrame:
        '''Return the dataframe at DatasetKeys.DATAFRAME, or raise an error if it doesn't exist.'''

        return self.dataframe



'''Holds cofirguration information about the dataset.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=protected-access

import abc
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from ...problem_def.config_component import (ConfigComponent, ConfigError,
                                             InvalidKeyError, InvalidValueError,
                                             MissingKeyError, ValidationErrors)
from ...problem_def.task import Task
from ...wrangler.dataset import Dataset, Metadata, RoleName, Column
from ...wrangler.constants import Defaults, ProblemDefKeys, ProblemDefKeySet


class DatasetConfigError(ConfigError):
    '''Base class for errors in DatasetConfig.'''


class UnknownConfigClassError(DatasetConfigError):
    '''Problem definition specified a dataset config that is not recognized.'''


class DatasetFileError(DatasetConfigError):
    '''The dataset file does not match information given in the problem definition.'''


class DatasetConfigTypeError(DatasetConfigError, TypeError):
    '''An object in the dataset config was not the right type.'''


class ParsingErrors(DatasetConfigError):
    '''At least one parsing error occured.'''
    def __init__(self, errors: List[DatasetConfigError]):
        super().__init__(f'At least one parsing error occured: {errors!r}')
        self.errors = errors


class DatasetConfig(ConfigComponent, metaclass=abc.ABCMeta):
    '''Holds confirguration information about the dataset.'''
    _roles: Dict[RoleName, List[Column]]
    _pos_labels: Dict[RoleName, Any]
    _forecasting_metadata: Dict[str, Any]
    _task: Optional[Task]

    def __init__(self,
                 clause: Dict[str, Any],
                 parents: Optional[List[str]] = None,
                 task: Optional[Task] = None,
                 **unused_kwargs):
        parents = self._add_parent(parents, ProblemDefKeys.CONFIG.value)
        super().__init__(clause=clause, parents=parents)
        self._roles = self._build_roles()
        self._pos_labels = self._build_pos_labels()
        self._forecasting_metadata = self._build_forecasting_metadata()
        self._task = task

    def __str__(self):
        return f'{{Index: {self.index_cols}, Target: {self.target}}}'

    def _build_roles(self) -> Dict[RoleName, List[Column]]:
        '''Build dict of column roles'''
        retval: Dict[RoleName, List[Column]] = {}
        for role in self._get_with_default(ProblemDefKeys.COL_ROLES, dflt={}):
            # this will throw a parsing error if the key does not match any known role name
            rolename = RoleName[role.upper()]

            col_names = self._get_with_default(ProblemDefKeys.COL_ROLES,
                                               role,
                                               ProblemDefKeys.COL_NAME,
                                               dflt=None)
            if col_names is not None:
                retval[rolename] = [Column(name=col_names)]
            # TODO(Merritt): make this more sensible

        return retval

    def _build_pos_labels(self) -> Dict[RoleName, Any]:
        '''Build dict of pos labels for binary classification.

        In the vast majority of cases, the only key will be RoleName.TARGET'''
        retval: Dict[RoleName, Any] = {}
        for role in self._get_with_default(ProblemDefKeys.COL_ROLES, dflt={}):
            rolename = RoleName[role.upper()]
            retval[rolename] = self._get_with_default(
                ProblemDefKeys.COL_ROLES, role, ProblemDefKeys.POS_LABEL, dflt=None)
        return retval

    def _build_forecasting_metadata(self) -> Dict[str, Any]:
        '''Build dict of additional metadata for forecasting problems.

        This includes 'horizon', 'input_size', and 'frequency'.
        '''
        retval: Dict[str, Any] = {}
        clause = self._get_with_default(ProblemDefKeys.FORECASTING, dflt={})
        if not isinstance(clause, dict):
            return retval
        for key, val in clause.items():
            retval[key] = val
        if ProblemDefKeys.FREQUENCY.value not in retval:
            retval[ProblemDefKeys.FREQUENCY.value] = Defaults.FREQUENCY

        # If step_size is not defined, set as equal to horizon
        if (ProblemDefKeys.STEP_SIZE.value not in retval
                and ProblemDefKeys.HORIZON.value in retval):
            retval[ProblemDefKeys.STEP_SIZE.value] = retval[ProblemDefKeys.HORIZON.value]

        return retval

    def validate(self, **kwargs) -> None:
        '''Check the problem definition dataset clause for errors.'''
        errors: List[ConfigError] = []

        dataset_keys = set(self._clause.keys())
        if not ProblemDefKeySet.DATASET.REQUIRED.issubset(dataset_keys):
            errors.append(MissingKeyError(
                'Required keys missing in dataset clause: '
                f'{ProblemDefKeySet.DATASET.REQUIRED.difference(dataset_keys)}'))

        if not dataset_keys.issubset(ProblemDefKeySet.DATASET.ALLOWED):
            errors.append(InvalidKeyError(
                'Invalid key(s) in dataset clause: '
                f'{dataset_keys.difference(ProblemDefKeySet.DATASET.ALLOWED)}'
            ))

        try:
            self._validate_roles()
        except ValidationErrors as err:
            errors.extend(err.errors)

        if self._exists(ProblemDefKeys.FORECASTING.value):
            data_keys = set(self._get(ProblemDefKeys.FORECASTING.value).keys())
            if not ProblemDefKeySet.DATASET.FORECASTING.REQUIRED.issubset(data_keys):
                errors.append(MissingKeyError(
                    'Required keys missing in forecasting metadata: '
                    f'{ProblemDefKeySet.DATASET.FORECASTING.REQUIRED.difference(data_keys)}'))

            if not data_keys.issubset(ProblemDefKeySet.DATASET.FORECASTING.ALLOWED):
                errors.append(InvalidKeyError(
                    'Invalid key(s) in forecasting metadata: '
                    f'{data_keys.difference(ProblemDefKeySet.DATASET.FORECASTING.ALLOWED)}'))

            freq_val = self._get_with_default(
                ProblemDefKeys.FORECASTING.value,
                ProblemDefKeys.FREQUENCY.value,
                dflt=Defaults.FREQUENCY)
            offset_strs = [x for x in dir(pd.offsets) if DatasetConfig._is_valid_offset(x)]
            allowed_freq = [getattr(pd.offsets, x)().freqstr for x in offset_strs]

            if freq_val not in allowed_freq:
                errors.append(InvalidValueError(
                    'Invalid value in forecasting metadata: '
                    f'{freq_val} is not a valid value for frequency. '
                    f'Allowed frequency values are: {allowed_freq}'))

            if ProblemDefKeys.HORIZON.value in data_keys:
                step_size = self.metadata.step_size or 0
                horizon = self._get(ProblemDefKeys.FORECASTING.value,
                                    ProblemDefKeys.HORIZON.value)
                if (step_size < 1) or (step_size > horizon):
                    errors.append(InvalidValueError(
                        'Invalid value in forecasting metadata: '
                        f'{step_size} is not a valid value for step_size. '
                        f'Value for step_size must be between 1 and {horizon}'
                    ))

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)

    def _validate_roles(self) -> None:
        if ProblemDefKeys.COL_ROLES.value not in set(self._clause.keys()):
            return

        errors: List[ConfigError] = []

        for role_name in self._get(ProblemDefKeys.COL_ROLES.value):
            role_keys = set(self._get(ProblemDefKeys.COL_ROLES.value,
                                      role_name).keys())
            if not ProblemDefKeySet.DATASET.ROLES.REQUIRED.issubset(role_keys):
                errors.append(MissingKeyError(
                    f'Required key(s) missing in role clause for role {role_name}: '
                    f'{ProblemDefKeySet.DATASET.ROLES.REQUIRED.difference(role_keys)}'))

            if not role_keys.issubset(ProblemDefKeySet.DATASET.ROLES.ALLOWED):
                errors.append(InvalidKeyError(
                    f'Invalid key(s) in role clause for role {role_name}: '
                    f'{role_keys.difference(ProblemDefKeySet.DATASET.ROLES.ALLOWED)}'
                ))

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)

    @staticmethod
    def _is_valid_offset(offset: str) -> bool:
        invalid_offsets = set(['BaseOffset', 'Tick', 'DateOffset', 'Easter'])
        obj = getattr(pd.offsets, offset)

        return not offset.startswith('_') and callable(obj) and offset not in invalid_offsets

    def cols_with_role(self, role: Union[str, RoleName]) -> List[Column]:
        '''Return a list of Columns that have the given role'''
        if isinstance(role, str):
            role = RoleName[role.upper()]
        if role not in self._roles:
            return []
        return self._roles[role]

    @property
    def target(self) -> Optional[Column]:
        '''Column with role "target".'''
        cols = self.cols_with_role(role=RoleName.TARGET)
        if len(cols) == 0:
            return None
        if len(cols) == 1:
            return cols[0]
        raise DatasetConfigError(f'Must have at most 1 target column, instead found {cols}')

    @property
    def roles(self) -> Dict[RoleName, List[Column]]:
        '''Full mapping of roles to lists of columns.'''
        return self._roles

    @property
    def index_cols(self) -> List[Column]:
        '''Column(s) containing indices for rows.'''
        return self.cols_with_role(role=RoleName.INDEX)

    @property
    def train_fraction(self) -> float:
        '''Config for the fraction of data to use for train.'''
        return self._get_with_default(
            ProblemDefKeys.SPLIT, ProblemDefKeys.TRAIN_FRAC, dflt=Defaults.SPLIT_FRACTION)

    def pos_label(self, role: RoleName) -> Optional[Any]:
        '''If this is a binary classification problem, return the positive label; otherwise None.'''
        return self._pos_labels[role]

    @property
    def forecasting(self) -> Optional[Dict[str, Any]]:
        '''Forecasting metadata'''
        retval = self._forecasting_metadata
        if not retval:
            return None
        return retval

    @property
    def metadata(self) -> Metadata:
        '''The metadata needed by models.'''
        task = None

        if self._task is not None:
            task = self._task.task_type

        return Metadata(roles=self._roles, pos_labels=self._pos_labels, task=task,
                        forecasting=self._forecasting_metadata)

    @abc.abstractmethod
    def load_train(self) -> Dataset:
        '''Load a Dataset from the information in the problem description'''

    @abc.abstractmethod
    def load_test(self) -> Optional[Dataset]:
        '''Load test data if it exists, otherwise None'''

    @abc.abstractmethod
    def dataset(self, data: Any, **kwargs) -> Dataset:
        '''Load a Dataset object, by placing data at supplied key.'''


class D3MDatasetConfig(DatasetConfig):
    '''Holds information about a local D3M formatted dataset.'''

    def load_train(self) -> Dataset:
        raise NotImplementedError

    def load_test(self) -> Optional[Dataset]:
        '''Load test data if it exists, otherwise None'''
        raise NotImplementedError

    def dataset(self, data: Any, **kwargs) -> Dataset:
        raise NotImplementedError


class TensorFlowDatasetConfig(DatasetConfig):
    '''Holds information about a local TensorFlow dataset.'''

    def load_train(self) -> Dataset:
        raise NotImplementedError

    def load_test(self) -> Optional[Dataset]:
        '''Load test data if it exists, otherwise None'''
        raise NotImplementedError

    def dataset(self, data: Any, **kwargs) -> Dataset:
        raise NotImplementedError


class CloudDatasetConfig(DatasetConfig):
    '''Holds information about a tabular dataset stored in the cloud.'''

    def load_train(self) -> Dataset:
        raise NotImplementedError

    def load_test(self) -> Optional[Dataset]:
        '''Load test data if it exists, otherwise None'''
        raise NotImplementedError

    def dataset(self, data: Any, **kwargs) -> Dataset:
        raise NotImplementedError



'''Holds information about a local tabular dataset.'''
import logging
from pathlib import Path
from typing import Optional, Dict, Any, List, Union

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd
from scipy.io.arff import loadarff  # type: ignore[import]

from .impl.dataframe_dataset_config import DataframeDatasetConfig
from .impl.dataset_config import DatasetFileError
from .impl.dataset_format_catalog import DatasetFormat, DatasetFormatCatalog
from ..problem_def.task import Task, TaskType
from ..wrangler.constants import ProblemDefKeys, FileType
from ..wrangler.dataset import Dataset, DatasetKeys, RoleName


class LocalDatasetConfig(DataframeDatasetConfig):
    '''Holds information about a local tabular dataset.'''
    _ext: Optional[str] = None
    _train_path: Optional[Path] = None
    _test_path: Optional[Path] = None
    _static_path: Optional[Path] = None

    def __init__(self,
                 clause: Dict[str, Any],
                 parents: Optional[List[str]] = None,
                 task: Optional[Task] = None,
                 **unused_kwargs):
        super().__init__(clause=clause, parents=parents, task=task)

        # This may throw errors, which will caught by ProblemDefinition
        # possible error states:
        # train path key doesn't exist
        # train path does not point to a csv
        # column roles key does not point to a dictionary
        # any of the roles have invalid names
        # any of the roles do not point to a dictionary
        # any of the roles have an invalid col name xor an invalid col id

        train_path: Path = Path(self._get(ProblemDefKeys.TRAIN_PATH.value))
        train_path = train_path.expanduser().resolve()
        ext: str = train_path.suffix[1:]

        if ext not in FileType.list():
            raise DatasetFileError(
                f'Train path from the problem definition ({train_path})'
                f' does not point to a file with one of these extensions: {FileType.list()}'
            )

        self._ext = ext

        try:
            train_cols = self._load_dataframe(train_path)
            assert train_cols is not None
            self._train_cols = train_cols
        except FileNotFoundError as exc:
            raise DatasetFileError(f'train_path: "{train_path}" not found') from exc
        self._train_path = train_path

        if ProblemDefKeys.TEST_PATH.value in clause:
            test_path = Path(self._get(ProblemDefKeys.TEST_PATH)).expanduser().resolve()
            test_ext = test_path.suffix[1:]
            if test_ext != ext:
                raise DatasetFileError(
                    f'test data: "{test_path}" is not same file type as train data'
                    f'(Expected: .{ext})')

            self._test_path = test_path

        if ProblemDefKeys.STATIC_EXOGENOUS_PATH.value in clause:
            static_path = Path(
                self._get(
                    ProblemDefKeys.STATIC_EXOGENOUS_PATH)).expanduser().resolve()
            static_ext = static_path.suffix[1:]
            if static_ext != ext:
                raise DatasetFileError(
                    f'static data: "{static_path}" is not same file type as train data'
                    f'(Expected: .{ext})')

            self._static_path = static_path

        self._read_roles()

    def _load_dataframe(self, path: Path) -> Optional[pd.DataFrame]:
        '''Load the file at path based on self._ext'''
        retval = None

        if self._ext == FileType.CSV.value:
            retval = pd.read_csv(path)
        elif self._ext == FileType.ARFF.value:
            raw_data = loadarff(path)
            retval = pd.DataFrame(raw_data[0])

        if retval is None:
            return None
        return self._drop_unnamed_cols(retval)

    def _drop_unnamed_cols(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        '''There are a few potential ways to do this.
        We choose to drop all cols called 'Unnamed: [int of any length]',
        as those were likely unnamed in the actual csv.'''
        # TODO(Merritt): warn user if we encounter this under certain circumstances
        # (for ex: not in first column
        # or the col is literally named 'Unnamed: [num]' in the csv)
        drop_cols = dataframe.filter(regex=r'^Unnamed: \d+$').columns
        retval = dataframe.drop(labels=drop_cols, axis=1)
        return retval

    def _load_dataset(self, path: Path, path_type: str) -> Dataset:
        retval = Dataset(metadata=self.metadata)
        # TODO(Mujing/Merritt): move this try/except statement into _load_dataframe()
        try:
            data = self._load_dataframe(path=path)
        except FileNotFoundError as exc:
            raise DatasetFileError(f'{path_type}_path: "{path}" not found') from exc

        if data is None:
            raise DatasetFileError(f'Unsupported file type: {path}')

        retval.dataframe = data

        # if this is a forecasting problem, load static exogenous data or create an empty key for it
        if self._task and self._task.task_type == TaskType.FORECASTING:
            if self._static_path is None:
                retval[DatasetKeys.STATIC_EXOGENOUS.value] = None
                logging.debug('Forecasting problem has no static exogenous table.')
            else:
                try:
                    static = self._load_dataframe(self._static_path)
                except FileNotFoundError as exc:
                    raise DatasetFileError(
                        f'static_path "{self._static_path}" not found. '
                        'This error should only appear in forecasting problems.') from exc
                retval[DatasetKeys.STATIC_EXOGENOUS.value] = static
            assert DatasetKeys.STATIC_EXOGENOUS.value in retval, (
                'BUG: forecasting problem but no "static_exogenous" data')

        return retval

    def load_train(self) -> Dataset:
        assert self._train_path is not None, 'BUG: load() called when train_path is None'
        return self._load_dataset(
            path=self._train_path,
            path_type='train')

    def load_test(self) -> Optional[Dataset]:
        if self._test_path is None:
            return None

        retval = self._load_dataset(
            path=self._test_path,
            path_type='test')

        return self._prune_target(retval)

    def dataset(self,
                data: Any,
                key: Union[DatasetKeys, str] = DatasetKeys.DATAFRAME,
                cols: Optional[List[str]] = None,
                roles: Optional[List[Union[RoleName, str]]] = None,
                **kwargs) -> Dataset:
        '''Load a Dataset object, by placing data at the supplied key.

        As of 2023-12-04 we only know how to handle things we can turn into a pandas DataFrame.

        Args:
          :data: Dataframe or object that can be turned into one.
            If data is a Path or a str, will attempt
            to load the file it points to as a dataframe and put that in the dataset.
          :key: The key for the data in the dataset. Defaults to "dataframe".
          :cols: If set, only selects supplied column(s).
            Will take union with roles if both are specified.
          :roles: If set, only selects columns with supplied role(s).
            Will take union with cols if both are specified.
        '''

        if isinstance(data, (Path, str)):
            path = Path(data)
            data = self._load_dataframe(path=path)

        return super().dataset(data, key=key, cols=cols, roles=roles, **kwargs)


class LocalDatasetFormat(DatasetFormat):
    '''Used to load a local tabular dataset.'''
    _builder = LocalDatasetConfig
    _name = 'local'
    _tags = {}


def register(catalog: DatasetFormatCatalog, *args, **kwargs):
    '''Register all the objects in this file.'''
    catalog.register(LocalDatasetFormat(*args, **kwargs))



'''Tests for DatasetConfig and DataseConfigFactory'''
# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path
from typing import Any, Optional

import pytest

from ...catalog.catalog import CatalogLookupError
from ...problem_def.config_component import ValidationErrors
from ...wrangler.dataset import Dataset, RoleName
from ..local import LocalDatasetConfig
from .dataset_format_catalog import DatasetFormatCatalogAuto
from .dataset_config import DatasetConfig, DatasetFileError

# pylint: disable=missing-function-docstring,duplicate-code,missing-class-docstring


class FakeDatasetConfig(DatasetConfig):

    def load_train(self) -> Dataset:
        return Dataset(metadata=self.metadata)

    def load_test(self) -> Optional[Dataset]:
        return Dataset(metadata=self.metadata)

    def dataset(self, data: Any, **kwargs) -> Dataset:
        return Dataset(metadata=self.metadata)


def valid_csv(filename: Optional[str] = None) -> str:
    '''Returns a path (in the form of a string) to a valid csv file.'''
    module_path = Path(__file__).parents[3]
    path = module_path / 'examples' / 'classification' / (filename or 'credit.csv')
    return str(path)


SUNNY_DATASET_CONFIG = {
    'config': 'local',
}


def test_dataset_format_catalog_sunny_day():
    clause = SUNNY_DATASET_CONFIG.copy()
    clause['train_path'] = valid_csv()

    config = DatasetFormatCatalogAuto().build(clause=clause)
    assert isinstance(config, LocalDatasetConfig)


NOT_FOUND_DATASET_CONFIG = {
    'config': 'non-existent',
}


def test_config_factory_not_found():
    with pytest.raises(CatalogLookupError, match=r'non-existent'):
        _ = DatasetFormatCatalogAuto().build(clause=NOT_FOUND_DATASET_CONFIG)


def test_config_factory_default():
    clause = {}
    clause['train_path'] = valid_csv()
    config = DatasetFormatCatalogAuto().build(clause=clause)
    assert isinstance(config, LocalDatasetConfig)


SUNNY_DATASET = {
    'config': 'local',
    'column_roles': {
        'target': {
            'name': 'class'
        }
    }
}


def test_dataset_sunny_day():
    clause = SUNNY_DATASET.copy()
    clause['train_path'] = valid_csv()

    config = DatasetFormatCatalogAuto().build(clause=clause)
    assert len(config.roles[RoleName.ATTRIBUTE]) == 20


TRAIN_PATH_MISSING = {
    'config': 'local'
}


def test_train_path_missing():
    with pytest.raises(KeyError, match=r'[Tt]rain'):
        _ = DatasetFormatCatalogAuto().build(clause=TRAIN_PATH_MISSING)


TRAIN_FILE_MISSING = {
    'config': 'local',
    'train_path': '/tmp/path/to/nonexistent_file.csv'
}


def test_train_file_missing():
    with pytest.raises(DatasetFileError, match=r'nonexistent'):
        _ = DatasetFormatCatalogAuto().build(clause=TRAIN_FILE_MISSING)


ROLE_WITH_NO_COL = {
    'config': 'local',
    'column_roles': {
        'target': {}
    }
}


def test_role_with_no_col():
    clause = ROLE_WITH_NO_COL.copy()
    clause['train_path'] = valid_csv()

    with pytest.raises(ValidationErrors, match=r'[Nn]ame'):
        DatasetFormatCatalogAuto().build(clause=clause).validate()


INVALID_KEY = {
    'config': 'local',
    'hamster': {
        'gerbil': 'gerbil',
    },
    'column_roles': {
        'target': {
            'name': 'class'
        },
        'index': {
            'id': 0
        }
    }
}


def test_dataset_invalid_key() -> None:
    clause = INVALID_KEY.copy()
    clause['train_path'] = valid_csv()

    dut = DatasetFormatCatalogAuto().build(clause=clause)

    with pytest.raises(ValidationErrors, match='hamster'):
        dut.validate()


POS_LABEL_DATASET = {
    'column_roles': {
        'target': {
            'name': 'a',
            'pos_label': 'good'
        }
    }
}


def test_pos_label_sunny_day() -> None:
    dut = FakeDatasetConfig(POS_LABEL_DATASET)
    assert 'good' == dut.load_train().metadata.pos_labels[RoleName.TARGET]


def test_pos_label_missing() -> None:
    clause = {
        'column_roles': {
            'target': {
                'name': 'class'
            }
        }
    }

    dut = FakeDatasetConfig(clause=clause)
    assert dut.pos_label(RoleName.TARGET) is None



'''Tests for single_fold_splitter.py'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd


from ..problem_def.cross_validation_config import CrossValidationConfig
from ..wrangler.dataset import Dataset, Metadata, RoleName, Column
from .single_fold_splitter import SingleFoldSplitter

# pylint: disable=missing-function-docstring, duplicate-code


def test_split_default_hyperparams() -> None:
    dataframe = pd.DataFrame({
        'a': range(1, 1001),
        'b': range(1001, 2001),
        'c': range(2001, 3001)})
    data = Dataset(metadata=Metadata(
        roles={
            RoleName.TARGET: [Column('c')]
        }
    ))
    data.dataframe = dataframe
    dut = SingleFoldSplitter(cv_config=CrossValidationConfig({}))

    got = dut.split(dataset=data)

    # default train_frac: 0.8
    got = dut.split(dataset=data)
    assert got.ground_truth is not None
    assert got.ground_truth.ground_truth.shape == (200, 1)
    assert len(got.folds) == 1
    assert got.folds[0].train.dataframe.shape == (800, 3)
    assert got.folds[0].validate.dataframe.shape == (200, 2)


def test_split_override_hyperparams() -> None:
    dataframe = pd.DataFrame({
        'a': range(1, 1001),
        'b': range(1001, 2001),
        'c': range(2001, 3001)})
    data = Dataset(metadata=Metadata(
        roles={
            RoleName.TARGET: [Column('c')]
        }
    ))
    data.dataframe = dataframe
    dut = SingleFoldSplitter(cv_config=CrossValidationConfig({}))

    got = dut.split(dataset=data, train_frac=0.6)
    assert got.ground_truth is not None
    assert got.ground_truth.ground_truth.shape == (400, 1)
    assert len(got.folds) == 1
    assert got.folds[0].train.dataframe.shape == (600, 3)
    assert got.folds[0].validate.dataframe.shape == (400, 2)



'''Holds an instance of a model.'''
import abc
from typing import Any
from typing_extensions import Protocol

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...wrangler.dataset import Dataset

from .algorithm import Algorithm


class Error(Exception):
    '''Base error class for AlgorithmInstance.'''


class DatasetError(Error):
    '''Something is malformed about the dataset.'''


class AlgorithmInstance(metaclass=abc.ABCMeta):
    '''Holds an instance of an algorithm as made by the instantiator.'''
    _algorithm: Algorithm

    def __init__(self, parent: Algorithm):
        self._algorithm = parent

    @abc.abstractmethod
    def predict(self, dataset: Dataset) -> Dataset:
        '''Apply model to input dataset to create output.

        This may require that the model is fit (self.trained == True) before it is called.
        '''

    @property
    def catalog_name(self) -> str:
        '''The catalog name of our algorithm.'''
        return self._algorithm.name


class Constructor(Protocol):
    '''Match the signature of a generic model constructor.'''
    def __call__(self, **kwargs: Any) -> Any:
        ...


def not_implemented(**kwargs):
    '''A constructor must be specified in a subclass'''
    raise NotImplementedError('Implementation must specify a model _constructor')



'''Wrappers of metrics from sklearn, currently only contains roc_auc_score.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Callable, Dict, List, Optional

import numpy as np
import sklearn.metrics  # type: ignore[import]
from sklearn.metrics import mean_squared_error  # type: ignore[import]

from ..problem_def.task import TaskType
from ..wrangler.constants import ProblemDefKeys
from ..wrangler.dataset import Dataset, DatasetKeyError, RoleName
from .impl.metric import MetricInvalidDatasetError, SupervisedMetric
from .impl.metric_catalog import MetricCatalog


class SklearnMetric(SupervisedMetric):
    '''Wrapper for sklean.metrics'''
    _metric: Optional[Callable[..., float]] = None

    def __init__(self,
                 name: str,
                 metric: Optional[Callable[..., float]] = None,
                 tasks: Optional[List[TaskType]] = None,
                 high: bool = True,
                 uses_proba: bool = False,
                 needs_pos_label: bool = False):
        # TODO(Merritt): high, uses_proba, needs_pos_label should be put in tags
        if metric is None:
            metric = getattr(sklearn.metrics, name, None)
        self._metric = metric
        if tasks is None:
            tasks = []
        self._tags = {
            ProblemDefKeys.TASK_TYPE.value: [t.name.lower() for t in tasks],
        }
        self._needs_pos_label = needs_pos_label
        self._high = high
        self._uses_proba = uses_proba
        super().__init__(name=name)

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        # TODO(Merritt): should be able to accept hyperparams from problem def
        ground_truth = self.validate_input(pred, ground_truth)
        # validate_input is guaranteed to return ground truth Dataset rather than None.
        assert ground_truth.metadata.target is not None
        ground_truth_for_sklearn = list(
            ground_truth.ground_truth[ground_truth.metadata.target.name])
        assert pred.metadata.target is not None, (
            'BUG: calculate() called with no target info in metadata'
        )

        if self.uses_proba:
            try:
                pred_df = pred.probabilities
            except DatasetKeyError:
                # We are not using 'from' because this error will be displayed
                #   alongside metric scores as a explanation for why this score
                #   cannot be calculated, thus we want it to be succint.
                raise MetricInvalidDatasetError(   # pylint: disable=raise-missing-from
                    'No class probabilities output.')
        else:
            pred_df = pred.predictions

        predictions_for_sklearn = list(
            pred_df[pred.metadata.target.name])

        assert self._metric is not None

        if RoleName.TARGET in pred.metadata.pos_labels and self._needs_pos_label:
            return self._metric(
                ground_truth_for_sklearn,
                predictions_for_sklearn,
                pos_label=pred.metadata.pos_labels[RoleName.TARGET])

        return self._metric(ground_truth_for_sklearn, predictions_for_sklearn)


def root_mean_squared_error(*args, **kwargs):
    '''Root Mean Squared Error'''
    return np.sqrt(mean_squared_error(*args, **kwargs))


METRICS: List[Dict[str, Any]] = [
    {
        'name': 'roc_auc_score',
        'tasks': [TaskType.BINARY_CLASSIFICATION],
        'uses_proba': True,
    },
    {
        'name': 'accuracy_score',
        'tasks': [TaskType.BINARY_CLASSIFICATION, TaskType.MULTICLASS_CLASSIFICATION],
    },
    {
        'name': 'f1_score',
        'high': True,
        'tasks': [TaskType.BINARY_CLASSIFICATION],
        'needs_pos_label': True,
    },
    {
        'name': 'mean_squared_error',
        'tasks': [TaskType.FORECASTING, TaskType.REGRESSION],
        'high': False,
    },
    {
        'name': 'root_mean_squared_error',
        'metric': root_mean_squared_error,
        'tasks': [TaskType.FORECASTING, TaskType.REGRESSION],
        'high': False,
    },
    {
        'name': 'mean_absolute_error',
        'tasks': [TaskType.FORECASTING, TaskType.REGRESSION],
        'high': False,
    },
    {
        'name': 'r2_score',
        'high': True,
        'tasks': [TaskType.REGRESSION],
    },
]


def register(catalog: MetricCatalog):
    '''Register all the metrics in this file.'''
    for met in METRICS:
        catalog.register(SklearnMetric(**met))



'''Tests for the Dataset class'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd
import pytest

from .dataset import Dataset, Metadata, RoleName, Column
from .dataset import DatasetKeys, DatasetKeyError, DatasetValueError

# pylint: disable=missing-function-docstring,pointless-statement


def test_output():
    roles = {RoleName.TIME: Column('bug')}
    dut = Dataset(metadata=Metadata(roles=roles))
    dut['some_key'] = 'some_val'
    got = dut.output().roles
    assert Column('bug') == got[RoleName.TIME]
    with pytest.raises(KeyError):
        got['some_key']


def test_output_overrides():
    roles = {RoleName.TIME: Column('bug')}
    roles2 = {RoleName.TIME: Column('bugs')}
    dut = Dataset(metadata=Metadata(roles=roles))
    dut['some_key'] = 'some_val'
    got = dut.output(override_metadata=Metadata(roles=roles2))
    assert Column('bugs') == got.roles[RoleName.TIME]
    with pytest.raises(KeyError):
        got['some_key']


def test_output_overrides_role():
    roles = {RoleName.TIME: Column('bug')}
    roles2 = {RoleName.TIME: Column('bugs')}
    dut = Dataset(metadata=Metadata(roles=roles, forecasting={'horizon': 10}))
    dut['some_key'] = 'some_val'
    got = dut.output(override_metadata=dut.metadata.override_roles(roles=roles2))
    assert Column('bugs') == got.roles[RoleName.TIME]
    assert got.metadata.horizon == 10


def test_get_dataframe():
    dut = Dataset()
    dataframe = pd.DataFrame({'bug': [1, 2], 'bug2': ['a', 'b']})
    dut.dataframe = dataframe

    assert 'bug' == dut.get_dataframe().columns[0]


def test_df_setters_and_getters() -> None:
    dut = Dataset()
    df1 = pd.DataFrame({'a': [1]})
    df2 = pd.DataFrame({'b': [2]})
    df3 = pd.DataFrame({'c': [3]})
    dut.dataframe = df1
    dut.ground_truth = df2
    dut.predictions = df3
    pd.testing.assert_frame_equal(df1, dut.dataframe)
    pd.testing.assert_frame_equal(df2, dut.ground_truth)
    pd.testing.assert_frame_equal(df3, dut.predictions)


def test_df_getters_proper_errors() -> None:
    dut = Dataset()
    with pytest.raises(DatasetKeyError):
        _ = dut.dataframe

    dut[DatasetKeys.DATAFRAME.value] = 'hamster'
    with pytest.raises(DatasetValueError, match='hamster'):
        _ = dut.dataframe

    dut[DatasetKeys.GROUND_TRUTH.value] = 'gerbil'
    with pytest.raises(DatasetValueError, match='gerbil'):
        _ = dut.ground_truth

    dut[DatasetKeys.PREDICTIONS.value] = 'skink'
    with pytest.raises(DatasetValueError, match='skink'):
        _ = dut.predictions
